"""
Functionality for reading Radarsat (RS2 and RCM) data into a SICD model.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Khanh Ho", "Wade Schwartzkopf", "Nathan Bombaci")


import logging
import re
import os
from datetime import datetime
from xml.etree import ElementTree
from typing import Tuple, List, Sequence, Union, Optional

import numpy
from scipy.interpolate import RectBivariateSpline
from numpy.polynomial import polynomial
from scipy.constants import speed_of_light

from sarpy.geometry.geocoords import geodetic_to_ecf

import sarpy._extensions
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.other_nitf import ComplexNITFReader
from sarpy.io.complex.sicd_elements.blocks import Poly1DType, Poly2DType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType, \
    RadarModeType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType
from sarpy.io.complex.sicd_elements.GeoData import GeoDataType, SCPType
from sarpy.io.complex.sicd_elements.Position import PositionType, XYZPolyType
from sarpy.io.complex.sicd_elements.Grid import GridType, DirParamType, WgtTypeType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType, \
    WaveformParametersType, ChanParametersType, TxStepType
from sarpy.io.complex.sicd_elements.Timeline import TimelineType, IPPSetType
from sarpy.io.complex.sicd_elements.ImageFormation import ImageFormationType, RcvChanProcType
from sarpy.io.complex.sicd_elements.RMA import RMAType, INCAType
from sarpy.io.complex.sicd_elements.SCPCOA import SCPCOAType
from sarpy.io.complex.sicd_elements.Radiometric import RadiometricType, NoiseLevelType_
from sarpy.io.complex.utils import fit_time_coa_polynomial, fit_position_xvalidation

from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.data_segment import DataSegment
from sarpy.io.general.tiff import NativeTiffDataSegment
from sarpy.io.general.utils import get_seconds, parse_timestring, is_file_like

logger = logging.getLogger(__name__)

_unhandled_generation_text = 'Unhandled generation `{}`'


############
# Helper functions

def _parse_xml(file_name: str, without_ns: bool = False) -> ElementTree.Element:
    if without_ns:
        with open(file_name, 'rb') as fi:
            xml_string = fi.read()
        # Remove the (first) default namespace definition (xmlns="http://some/namespace") and parse
        return ElementTree.fromstring(re.sub(b'\\sxmlns="[^"]+"', b'', xml_string, count=1))
    else:
        return ElementTree.parse(file_name).getroot()


def _format_class_str(class_str: str) -> str:
    if 'UNCLASS' in class_str:
        return 'UNCLASSIFIED'
    else:
        return class_str


def load_addin():
    """Check for a Radarsat addin module"""
    try:
        from sarpy.io.complex import radarsat_addin
        return radarsat_addin
    except ImportError:
        pass

    eps = sarpy._extensions.entry_points(group='sarpy.io.complex.radarsat')
    if not eps:
        return None
    if len(eps) > 1:
        raise RuntimeError("More than one Radarsat addin found")
    return list(eps)[0].load()


def _validate_segment_and_sicd(
        the_sicd: SICDType,
        data_segment: DataSegment,
        name: str,
        the_file: str):
    """
    Check that chipper and sicd are compatible.

    Parameters
    ----------
    the_sicd : SICDType
    data_segment : DataSegment
    name : str
    the_file : str

    Returns
    -------
    None
    """

    sicd_data_size = (the_sicd.ImageData.NumRows, the_sicd.ImageData.NumCols)
    segment_data_size = data_segment.formatted_shape
    if sicd_data_size != segment_data_size:
        raise ValueError(
            'The {} data segment construction for file {}\n'
            'got incompatible sicd size `{}` and segment size `{}`'.format(
                name, the_file, sicd_data_size, segment_data_size))


def _construct_tiff_segment(
        the_sicd: SICDType,
        the_file: str,
        reverse_axes: Union[None, int, Sequence[int]] = None,
        transpose_axes: Union[None, Tuple[int, ...]] = None):
    """

    Parameters
    ----------
    the_sicd : SICDType
    the_file : str
    reverse_axes : None|Tuple[int, ...]
    transpose_axes : None|Tuple[int, ...]

    Returns
    -------
    NativeTiffDataSegment
    """

    segment = NativeTiffDataSegment(the_file, reverse_axes=reverse_axes, transpose_axes=transpose_axes)
    _validate_segment_and_sicd(the_sicd, segment, 'tiff', the_file)
    return segment


def _construct_single_nitf_segment(
        the_sicd: SICDType,
        the_file: str,
        reverse_axes: Optional[Sequence[int]],
        transpose_axes: Optional[Tuple[int, ...]]) -> Tuple[ComplexNITFReader, DataSegment]:
    """

    Parameters
    ----------
    the_sicd : SICDType
    the_file : str
    reverse_axes : None|Sequence[int]
    transpose_axes : None|Tuple[int, ...]

    Returns
    -------
    reader: ComplexNITFReader
    data_segment: DataSegment
    """

    if transpose_axes is not None:
        transpose_axes = transpose_axes[:2]
    reader = ComplexNITFReader(the_file, reverse_axes=reverse_axes, transpose_axes=transpose_axes)
    data_segment = reader.data_segment
    if not isinstance(data_segment, DataSegment):
        raise ValueError(
            'The SLC data for a single polarmetric band was provided '
            'in a NITF file which has more than a single complex band.')
    _validate_segment_and_sicd(the_sicd, data_segment, 'Single NITF', the_file)
    return reader, data_segment


def _construct_multiple_nitf_segment(
        the_sicds: List[SICDType],
        the_file: str,
        reverse_axes: Optional[Sequence[int]],
        transpose_axes: Optional[Tuple[int, ...]]) -> Tuple[ComplexNITFReader, Tuple[DataSegment, ...]]:
    """

    Parameters
    ----------
    the_sicds : List[SICDType]
    the_file : str
    reverse_axes : None|Sequence[int]
    transpose_axes : None|Tuple[int, ...]

    Returns
    -------
    reader: ComplexNITFReader
    data_segment: Tuple[DataSegment, ...]
    """

    if transpose_axes is not None:
        transpose_axes = transpose_axes[:2]
    reader = ComplexNITFReader(the_file, reverse_axes=reverse_axes, transpose_axes=transpose_axes)
    data_segment = reader.get_data_segment_as_tuple()
    if len(data_segment) != len(the_sicds):
        raise ValueError(
            'The SLC data for {} polarmetric bands was provided '
            'in a NITF file which has {} single complex band.'.format(len(the_sicds), len(data_segment)))
    for i, (the_sicd, segment) in enumerate(zip(the_sicds, data_segment)):
        _validate_segment_and_sicd(the_sicd, segment, 'NITF band {}'.format(i), the_file)
    return reader, data_segment


##############
# Class for meta-data interpretation

class RadarSatDetails(object):
    """
    Class for interpreting RadarSat-2 and RadarSat Constellation Mission (RCM)
    metadata files, and creating the corresponding sicd structure(s).
    """

    __slots__ = (
        '_file_name', '_directory_name', '_satellite', '_root_node', '_beams', '_bursts',
        '_num_lines_processed', '_polarizations',
        '_x_spline', '_y_spline', '_z_spline',
        '_state_time', '_state_position', '_state_velocity')

    def __init__(self, file_name: str):
        """

        Parameters
        ----------
        file_name : str
        """

        self._beams = None
        self._bursts = None
        self._num_lines_processed = None
        self._polarizations = None
        self._x_spline = None
        self._y_spline = None
        self._z_spline = None
        self._state_time = None
        self._state_position = None
        self._state_velocity = None

        if os.path.isdir(file_name):  # it is the directory - point it at the product.xml file
            for t_file_name in [
                    os.path.join(file_name, 'product.xml'),
                    os.path.join(file_name, 'metadata', 'product.xml')]:
                if os.path.exists(t_file_name):
                    file_name = t_file_name
                    break
        if not os.path.isfile(file_name):
            raise SarpyIOError('path {} does not exist or is not a file'.format(file_name))
        if os.path.split(file_name)[1] != 'product.xml':
            raise SarpyIOError(
                'The radarsat or rcm file is expected to be named product.xml,\n\t'
                'got path {}'.format(file_name))

        self._file_name = file_name
        root_node = _parse_xml(file_name, without_ns=True)

        sat_node = root_node.find('./sourceAttributes/satellite')
        satellite = 'None' if sat_node is None else sat_node.text.upper()
        product_node = root_node.find(
            './imageGenerationParameters/generalProcessingInformation/productType')
        product_type = 'None' if product_node is None else product_node.text.upper()
        if not ((satellite == 'RADARSAT-2' or satellite.startswith('RCM')) and product_type == 'SLC'):
            raise SarpyIOError(
                'File {} does not appear to be an SLC product\n\t'
                'for a RADARSAT-2 or RCM mission.'.format(file_name))

        self._root_node = root_node
        self._satellite = satellite
        absolute_path = os.path.abspath(self._file_name)
        parent_dir, _ = os.path.split(absolute_path)
        if self.generation == 'RS2':
            self._directory_name = parent_dir
        else:
            self._directory_name, _ = os.path.split(parent_dir)

        self._build_location_spline()
        self._parse_state_vectors()
        self._extract_beams_and_bursts()

    @property
    def file_name(self) -> str:
        """
        str: the file name
        """

        return self._file_name

    @property
    def directory_name(self) -> str:
        """
        str: the package directory name
        """

        return self._directory_name

    @property
    def satellite(self) -> str:
        """
        str: the satellite name
        """

        return self._satellite

    @property
    def generation(self) -> str:
        """
        str: RS2 or RCM
        """

        if self._satellite == 'RADARSAT-2':
            return 'RS2'
        else:
            return 'RCM'

    @property
    def pass_direction(self) -> str:
        """
        str: The pass direction
        """

        return self._find('./sourceAttributes/orbitAndAttitude/orbitInformation/passDirection').text

    def get_symmetry(self) -> Tuple[Optional[Tuple[int, ...]], Optional[Tuple[int, ...]]]:
        """
        Get the symmetry transform information.

        Returns
        -------
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]
        """

        look_dir = self._find('./sourceAttributes/radarParameters/antennaPointing').text.upper()[0]
        if self.generation == 'RS2':
            line_order = self._find('./imageAttributes/rasterAttributes/lineTimeOrdering').text.upper()
            sample_order = self._find('./imageAttributes/rasterAttributes/pixelTimeOrdering').text.upper()
        else:
            line_order = self._find('./imageReferenceAttributes/rasterAttributes/lineTimeOrdering').text.upper()
            sample_order = self._find('./imageReferenceAttributes/rasterAttributes/pixelTimeOrdering').text.upper()
        reverse_cols = not (
                (line_order == 'DECREASING' and look_dir == 'L') or
                (line_order != 'DECREASING' and look_dir != 'L'))
        reverse_axes = []
        if reverse_cols:
            reverse_axes.append(0)
        if sample_order == 'DECREASING':
            reverse_axes.append(1)
        reverse_axes = tuple(reverse_axes) if len(reverse_axes) > 0 else None
        transpose_axes = (1, 0, 2)
        return reverse_axes, transpose_axes

    def _find(self, tag: str) -> ElementTree.Element:
        return self._root_node.find(tag)

    def _findall(self, tag: str) -> List[ElementTree.Element]:
        return self._root_node.findall(tag)

    def _get_tiepoint_nodes(self) -> List[ElementTree.Element]:
        """
        Fetch the tie point nodes.

        Returns
        -------
        List[ElementTree.Element]
        """

        return self._findall('./imageAttributes/geographicInformation/geolocationGrid/imageTiePoint')

    def _build_location_spline(self) -> None:
        """
        Populates the three (line, sample) -> location coordinate splines. This
        should be done once for all images.

        Returns
        -------
        None
        """

        if self.generation == 'RS2':
            tie_points = self._findall('./imageAttributes/geographicInformation/geolocationGrid/imageTiePoint')
        elif self.generation == 'RCM':
            tie_points = self._findall('./imageReferenceAttributes/geographicInformation/geolocationGrid/imageTiePoint')
        else:
            raise ValueError('unexpected generation {}'.format(self.generation))

        lines = []
        samples = []
        llh_coords = numpy.zeros((len(tie_points), 3), dtype='float64')
        grid_row, grid_col = None, None
        for i, entry in enumerate(tie_points):
            img_coords = entry.find('./imageCoordinate')
            geo_coords = entry.find('./geodeticCoordinate')
            # parse lat/lon/hae
            llh_coords[i, :] = [
                float(geo_coords.find('./latitude').text),
                float(geo_coords.find('./longitude').text),
                float(geo_coords.find('./height').text)]

            # parse line/sample
            line = float(img_coords.find('./line').text)
            sample = float(img_coords.find('./pixel').text)
            # verify grid structure
            if i == 0:
                lines.append(line)
                samples.append(sample)
                grid_row = 0
                grid_col = 0
                continue

            if sample == samples[0]:
                # we are starting a new grid column
                grid_row += 1
                grid_col = 0
                lines.append(line)
            else:
                grid_col += 1
                if grid_row == 0:
                    samples.append(sample)

            # verify that the grid assumption is preserved
            if grid_col >= len(samples) or grid_row >= len(lines) or \
                    line != lines[grid_row] or sample != samples[grid_col]:
                logger.error(
                    'Failed parsing grid at\n\t'
                    'grid_col = {}\n\t'
                    'samples = {}\n\t'
                    'grid_row = {}\n\t'
                    'lines = {}\n\t'
                    'line={}, sample={}'.format(grid_col, samples, grid_row, lines, line, sample))
                raise ValueError('The grid assumption is invalid at imageTiePoint entry {}'.format(i))
        lines = numpy.array(lines, dtype='float64')
        samples = numpy.array(samples, dtype='float64')
        ecf_coords = geodetic_to_ecf(llh_coords)
        self._x_spline = RectBivariateSpline(
            lines, samples, numpy.reshape(ecf_coords[:, 0], (lines.size, samples.size)), kx=3, ky=3, s=0)
        self._y_spline = RectBivariateSpline(
            lines, samples, numpy.reshape(ecf_coords[:, 1], (lines.size, samples.size)), kx=3, ky=3, s=0)
        self._z_spline = RectBivariateSpline(
            lines, samples, numpy.reshape(ecf_coords[:, 2], (lines.size, samples.size)), kx=3, ky=3, s=0)

    def _get_image_location(self, line: Union[int, float], sample: Union[int, float]) -> numpy.ndarray:
        """
        Fetch the image location estimate based on the previously constructed splines.

        Parameters
        ----------
        line : int|float
            The RadarSat line number.
        sample : int|float
            The RadarSat sample number.

        Returns
        -------
        numpy.ndarray
        """

        return numpy.array(
            [float(self._x_spline.ev(line, sample)),
             float(self._y_spline.ev(line, sample)),
             float(self._z_spline.ev(line, sample))], dtype='float64')

    def _parse_state_vectors(self) -> None:
        """
        Parses the state vectors.

        Returns
        -------
        None
        """

        state_vectors = self._findall(
            './sourceAttributes/orbitAndAttitude/orbitInformation/stateVector')

        self._state_time = numpy.zeros((len(state_vectors), ), dtype='datetime64[us]')
        self._state_position = numpy.zeros((len(state_vectors), 3), dtype='float64')
        self._state_velocity = numpy.zeros((len(state_vectors), 3), dtype='float64')

        for i, state_vec in enumerate(state_vectors):
            self._state_time[i] = parse_timestring(
                state_vec.find('timeStamp').text, precision='us')
            self._state_position[i, :] = [
                float(state_vec.find('xPosition').text),
                float(state_vec.find('yPosition').text),
                float(state_vec.find('zPosition').text)]
            self._state_velocity[i, :] = [
                float(state_vec.find('xVelocity').text),
                float(state_vec.find('yVelocity').text),
                float(state_vec.find('zVelocity').text)]

    def _extract_beams_and_bursts(self) -> None:
        """
        Extract the beam and burst and polarization information.

        Returns
        -------
        None
        """

        radar_params = self._find('./sourceAttributes/radarParameters')
        self._beams = radar_params.find('./beams').text.strip().split()
        self._polarizations = radar_params.find('./polarizations').text.strip().split()
        if self.generation == 'RCM':
            image_attributes = self._findall('./sceneAttributes/imageAttributes')
            if 'burst' in image_attributes[0].attrib:
                self._bursts = [(entry.attrib['beam'], entry.attrib['burst']) for entry in image_attributes]
                num_lines_processed = 0
                self._bursts = []
                for entry in image_attributes:
                    self._bursts.append((entry.attrib['beam'], entry.attrib['burst']))
                    nlines = int(entry.find('./numLines').text)
                    line_offset = int(entry.find('./lineOffset').text)
                    num_lines_processed = max(num_lines_processed, nlines+line_offset)
                self._num_lines_processed = num_lines_processed

    def _get_sicd_radar_mode(self) -> RadarModeType:
        """
        Gets the RadarMode information.

        Returns
        -------
        RadarModeType
        """

        mode_id = self._find('./sourceAttributes/beamModeMnemonic').text
        beam_mode = self._find('./sourceAttributes/beamMode')
        acq_type = self._find('./sourceAttributes/radarParameters/acquisitionType')
        if (beam_mode is not None and beam_mode.text.upper().startswith("SPOTLIGHT")) \
                or (acq_type is not None and acq_type.text.upper().startswith("SPOTLIGHT")) \
                or 'SL' in mode_id:
            mode_type = 'SPOTLIGHT'
        elif mode_id.startswith('SC'):
            # ScanSAR modes
            mode_type = 'SPOTLIGHT'
        else:
            mode_type = 'STRIPMAP'
        return RadarModeType(ModeID=mode_id, ModeType=mode_type)

    def _get_sicd_collection_info(self, start_time: numpy.datetime64) -> Tuple[dict, CollectionInfoType]:
        """
        Gets the sicd CollectionInfo information.

        Parameters
        ----------
        start_time : numpy.datetime64

        Returns
        -------
        nitf : dict
            The NITF element dictionary
        collection_info : CollectionInfoType
        """

        radarsat_addin = load_addin()

        collector_name = self.satellite
        start_time_dt = start_time.astype(datetime)
        date_str = start_time_dt.strftime('%d%b%y').upper()
        nitf = {}
        if self.generation == 'RS2':
            classification = 'UNCLASSIFIED'
            core_name = '{}{}{}'.format(date_str, self.generation, self._find('./sourceAttributes/imageId').text)
        elif self.generation == 'RCM':
            class_str = self._find('./securityAttributes/securityClassification').text.upper()
            classification = _format_class_str(class_str) if radarsat_addin is None else \
                radarsat_addin.extract_radarsat_sec(nitf, class_str)
            core_name = '{}{}{}'.format(date_str, collector_name.replace('-', ''), start_time_dt.strftime('%H%M%S'))
        else:
            raise ValueError(_unhandled_generation_text.format(self.generation))

        return nitf, CollectionInfoType(
            Classification=classification,
            CollectorName=collector_name,
            CoreName=core_name,
            RadarMode=self._get_sicd_radar_mode(),
            CollectType='MONOSTATIC')

    def _get_sicd_image_creation(self) -> ImageCreationType:
        """
        Gets the ImageCreation metadata.

        Returns
        -------
        ImageCreationType
        """

        from sarpy.__about__ import __version__
        processing_info = self._find('./imageGenerationParameters/generalProcessingInformation')
        return ImageCreationType(
            Application=processing_info.find('softwareVersion').text,
            DateTime=processing_info.find('processingTime').text,
            Site=processing_info.find('processingFacility').text,
            Profile='sarpy {}'.format(__version__))

    def _get_sicd_position(self, start_time: numpy.datetime64) -> PositionType:
        """
        Gets the SICD Position definition, based on the given start time.

        Parameters
        ----------
        start_time : numpy.datetime64

        Returns
        -------
        PositionType
        """

        # convert to relative time for polynomial fitting
        T = numpy.array([get_seconds(entry, start_time, precision='us') for entry in self._state_time], dtype='float64')
        P_x, P_y, P_z = fit_position_xvalidation(T, self._state_position, self._state_velocity, max_degree=8)
        return PositionType(ARPPoly=XYZPolyType(X=P_x, Y=P_y, Z=P_z))

    @staticmethod
    def _parse_polarization(str_in: str) -> Tuple[str, str]:
        """
        Parses the Radarsat polarization string into it's two SICD components.

        Parameters
        ----------
        str_in : str

        Returns
        -------
        (str, str)
        """

        if len(str_in) != 2:
            raise ValueError('Got input string of unexpected length {}'.format(str_in))

        tx_pol = 'RHC' if str_in[0] == 'C' else str_in[0]
        rcv_pol = 'RHC' if str_in[1] == 'C' else str_in[1]  # probably only H/V
        return tx_pol, rcv_pol

    def _get_sicd_polarizations(self) -> Tuple[List[str], List[str]]:
        tx_pols = []
        tx_rcv_pols = []
        for entry in self._polarizations:
            tx_pol, rcv_pol = self._parse_polarization(entry)
            if tx_pol not in tx_pols:
                tx_pols.append(tx_pol)
            tx_rcv_pols.append('{}:{}'.format(tx_pol, rcv_pol))
        return tx_pols, tx_rcv_pols

    def _get_side_of_track(self) -> str:
        """
        Gets the sicd side of track.

        Returns
        -------
        str
        """

        return self._find('./sourceAttributes/radarParameters/antennaPointing').text[0].upper()

    def _get_regular_sicd(self) -> Tuple[List[SICDType], List[str]]:
        """
        Gets the SICD collection. This will return one SICD per polarimetric
        collection. It will also return the data file(s). This is only applicable
        for non-ScanSAR collects.

        Returns
        -------
        sicds: List[SICDType]
        files: List[str]
        """

        def get_image_and_geo_data() -> Tuple[ImageDataType, GeoDataType]:
            if self.generation == 'RS2':
                pixel_type = 'RE16I_IM16I'
                cols = int(self._find('./imageAttributes/rasterAttributes/numberOfLines').text)
                rows = int(self._find('./imageAttributes/rasterAttributes/numberOfSamplesPerLine').text)
            elif self.generation == 'RCM':
                cols = int(self._find('./sceneAttributes/imageAttributes/numLines').text)
                rows = int(self._find('./sceneAttributes/imageAttributes/samplesPerLine').text)
                bits_per_sample = self._find('./imageReferenceAttributes/rasterAttributes/bitsPerSample').text
                if bits_per_sample == '32':
                    pixel_type = 'RE32F_IM32F'
                elif bits_per_sample == '16':
                    pixel_type = 'RE16I_IM16I'
                else:
                    raise ValueError('Got unhandled bites per sample {}'.format(bits_per_sample))
            else:
                raise ValueError(_unhandled_generation_text.format(self.generation))
            scp_rows = int(0.5*rows)
            scp_cols = int(0.5*cols)
            scp_ecf = self._get_image_location(scp_cols, scp_rows)
            im_data = ImageDataType(
                NumRows=rows, NumCols=cols, FirstRow=0, FirstCol=0, PixelType=pixel_type,
                FullImage=(rows, cols), SCPPixel=(scp_rows, scp_cols))
            t_geo_data = GeoDataType(SCP=SCPType(ECF=scp_ecf))
            return im_data, t_geo_data

        def get_grid_row() -> DirParamType:
            if self.generation == 'RS2':
                row_ss = float(self._find('./imageAttributes/rasterAttributes/sampledPixelSpacing').text)
                row_irbw = 2*float(self._find('./imageGenerationParameters'
                                              '/sarProcessingInformation'
                                              '/totalProcessedRangeBandwidth').text)/speed_of_light
            elif self.generation == 'RCM':
                row_ss = float(self._find('./imageReferenceAttributes/rasterAttributes/sampledPixelSpacing').text)
                row_irbw = 2*float(self._find('./sourceAttributes'
                                              '/radarParameters'
                                              '/pulseBandwidth').text)/speed_of_light
            else:
                raise ValueError(_unhandled_generation_text.format(self.generation))
            row_wgt_type = WgtTypeType(
                WindowName=self._find('./imageGenerationParameters'
                                      '/sarProcessingInformation'
                                      '/rangeWindow/windowName').text.upper())
            if row_wgt_type.WindowName == 'KAISER':
                row_wgt_type.Parameters = {
                    'BETA': self._find('./imageGenerationParameters'
                                       '/sarProcessingInformation'
                                       '/rangeWindow/windowCoefficient').text}
            return DirParamType(
                SS=row_ss, ImpRespBW=row_irbw, Sgn=-1, KCtr=2*center_frequency/speed_of_light,
                DeltaKCOAPoly=Poly2DType(Coefs=((0,),)), WgtType=row_wgt_type)

        def get_grid_col() -> DirParamType:
            az_win = self._find('./imageGenerationParameters/sarProcessingInformation/azimuthWindow')
            col_wgt_type = WgtTypeType(WindowName=az_win.find('./windowName').text.upper())
            if col_wgt_type.WindowName == 'KAISER':
                col_wgt_type.Parameters = {'BETA': az_win.find('./windowCoefficient').text}
            return DirParamType(Sgn=-1, KCtr=0, WgtType=col_wgt_type)

        def get_radar_collection() -> RadarCollectionType:
            radar_params = self._find('./sourceAttributes/radarParameters')
            # Ultrafine and spotlight modes have t pulses, otherwise just one.
            bandwidth_elements = sorted(radar_params.findall('pulseBandwidth'), key=lambda x: x.get('pulse'))
            pulse_length_elements = sorted(radar_params.findall('pulseLength'), key=lambda x: x.get('pulse'))
            adc_elements = sorted(radar_params.findall('adcSamplingRate'), key=lambda x: x.get('pulse'))
            samples_per_echo = float(radar_params.find('samplesPerEchoLine').text)
            wf_params = []
            bandwidths = numpy.empty((len(bandwidth_elements),), dtype=numpy.float64)
            for j, (bwe, ple, adce) in enumerate(zip(bandwidth_elements, pulse_length_elements, adc_elements)):
                bandwidths[j] = float(bwe.text)
                samp_rate = float(adce.text)
                wf_params.append(WaveformParametersType(index=j,
                                                        TxRFBandwidth=float(bwe.text),
                                                        TxPulseLength=float(ple.text),
                                                        ADCSampleRate=samp_rate,
                                                        RcvWindowLength=samples_per_echo / samp_rate,
                                                        RcvDemodType='CHIRP',
                                                        RcvFMRate=0))
            tot_bw = numpy.sum(bandwidths)
            tx_freq = (center_frequency-0.5*tot_bw, center_frequency+0.5*tot_bw)
            t_radar_collection = RadarCollectionType(TxFrequency=tx_freq, Waveform=wf_params)
            t_radar_collection.Waveform[0].TxFreqStart = tx_freq[0]
            for j in range(1, len(bandwidth_elements)):
                t_radar_collection.Waveform[j].TxFreqStart = t_radar_collection.Waveform[j - 1].TxFreqStart + \
                                                             t_radar_collection.Waveform[j - 1].TxRFBandwidth
            t_radar_collection.RcvChannels = [
                ChanParametersType(TxRcvPolarization=entry, index=j+1) for j, entry in enumerate(tx_rcv_pols)]
            if len(tx_pols) == 1:
                t_radar_collection.TxPolarization = tx_pols[0]
            else:
                t_radar_collection.TxPolarization = 'SEQUENCE'
                t_radar_collection.TxSequence = [
                    TxStepType(TxPolarization=entry, index=j+1) for j, entry in enumerate(tx_pols)]
            return t_radar_collection

        def get_timeline() -> TimelineType:
            pulse_parts = len(self._findall('./sourceAttributes/radarParameters/pulseBandwidth'))
            if self.generation == 'RS2':
                pulse_rep_freq = float(
                    self._find('./sourceAttributes/radarParameters/pulseRepetitionFrequency').text)
            elif self.generation == 'RCM':
                pulse_rep_freq = float(
                    self._find('./sourceAttributes/radarParameters/prfInformation/pulseRepetitionFrequency').text)
            else:
                raise ValueError(_unhandled_generation_text.format(self.generation))

            pulse_rep_freq *= pulse_parts
            if pulse_parts == 2 and collection_info.RadarMode.ModeType == 'STRIPMAP':
                # it's not completely clear why we need an additional factor of 2 for strip map
                pulse_rep_freq *= 2
            lines_processed = [
                float(entry.text) for entry in
                self._findall('./imageGenerationParameters/sarProcessingInformation/numberOfLinesProcessed')]
            duration = None
            ipp = None
            # there should be one entry of num_lines_processed for each transmit/receive polarization
            # and they should all be the same. Omit if this is not the case.
            if (len(lines_processed) == len(tx_rcv_pols)) and all(x == lines_processed[0] for x in lines_processed):
                num_lines_processed = lines_processed[0] * len(tx_pols)
                duration = num_lines_processed / pulse_rep_freq
                ipp = IPPSetType(
                    index=0, TStart=0, TEnd=duration, IPPStart=0, IPPEnd=int(num_lines_processed),
                    IPPPoly=Poly1DType(Coefs=(0, pulse_rep_freq)))
            return TimelineType(
                CollectStart=collect_start, CollectDuration=duration, IPP=[ipp, ])

        def get_image_formation() -> ImageFormationType:
            pulse_parts = len(self._findall('./sourceAttributes/radarParameters/pulseBandwidth'))
            return ImageFormationType(
                # PRFScaleFactor for either polarimetric or multi-step, but not both.
                RcvChanProc=RcvChanProcType(
                    NumChanProc=1, PRFScaleFactor=1./max(pulse_parts, len(tx_pols))),
                ImageFormAlgo='RMA',
                TStartProc=timeline.IPP[0].TStart,
                TEndProc=timeline.IPP[0].TEnd,
                TxFrequencyProc=(
                    radar_collection.TxFrequency.Min, radar_collection.TxFrequency.Max),
                STBeamComp='GLOBAL',
                ImageBeamComp='SV',
                AzAutofocus='NO',
                RgAutofocus='NO')

        def get_rma_adjust_grid() -> RMAType:
            # fetch all the things needed below
            # generation agnostic
            doppler_bandwidth = float(
                self._find('./imageGenerationParameters'
                           '/sarProcessingInformation'
                           '/totalProcessedAzimuthBandwidth').text)
            zero_dop_last_line = \
                parse_timestring(
                    self._find('./imageGenerationParameters'
                               '/sarProcessingInformation'
                               '/zeroDopplerTimeLastLine').text,
                    precision='us')
            zero_dop_first_line = parse_timestring(
                self._find('./imageGenerationParameters'
                           '/sarProcessingInformation'
                           '/zeroDopplerTimeFirstLine').text,
                precision='us')
            if self.generation == 'RS2':
                near_range = float(
                    self._find('./imageGenerationParameters'
                               '/sarProcessingInformation'
                               '/slantRangeNearEdge').text)

                doppler_rate_node = self._find('./imageGenerationParameters'
                                               '/dopplerRateValues')
                doppler_rate_coeffs = numpy.array(
                    [float(entry) for entry in doppler_rate_node.find('./dopplerRateValuesCoefficients').text.split()],
                    dtype='float64')

                doppler_centroid_node = self._find('./imageGenerationParameters'
                                                   '/dopplerCentroid')
            elif self.generation == 'RCM':
                near_range = float(
                    self._find('./sceneAttributes/imageAttributes/slantRangeNearEdge').text)

                doppler_rate_node = self._find('./dopplerRate'
                                               '/dopplerRateEstimate')
                doppler_rate_coeffs = numpy.array(
                    [float(entry) for entry in doppler_rate_node.find('./dopplerRateCoefficients').text.split()],
                    dtype='float64')

                doppler_centroid_node = self._find('./dopplerCentroid'
                                                   '/dopplerCentroidEstimate')
            else:
                raise ValueError(_unhandled_generation_text.format(self.generation))

            doppler_rate_ref_time = float(doppler_rate_node.find('./dopplerRateReferenceTime').text)
            doppler_cent_coeffs = numpy.array(
                [float(entry) for entry in
                 doppler_centroid_node.find('./dopplerCentroidCoefficients').text.split()],
                dtype='float64')
            doppler_cent_ref_time = float(
                doppler_centroid_node.find('./dopplerCentroidReferenceTime').text)
            doppler_cent_time_est = parse_timestring(
                doppler_centroid_node.find('./timeOfDopplerCentroidEstimate').text, precision='us')

            look = scpcoa.look
            if look > 0:
                # SideOfTrack == 'L'
                # we explicitly want negative time order
                if zero_dop_first_line < zero_dop_last_line:
                    zero_dop_first_line, zero_dop_last_line = zero_dop_last_line, zero_dop_first_line
            else:
                # we explicitly want positive time order
                if zero_dop_first_line > zero_dop_last_line:
                    zero_dop_first_line, zero_dop_last_line = zero_dop_last_line, zero_dop_first_line
            col_spacing_zd = get_seconds(zero_dop_last_line, zero_dop_first_line, precision='us') / \
                (image_data.NumCols - 1)
            # zero doppler time of SCP relative to collect start
            time_scp_zd = get_seconds(zero_dop_first_line, collect_start, precision='us') + \
                image_data.SCPPixel.Col * col_spacing_zd

            inca = INCAType(
                R_CA_SCP=near_range + (image_data.SCPPixel.Row * grid.Row.SS),
                FreqZero=center_frequency)
            # doppler rate calculations
            velocity = position.ARPPoly.derivative_eval(time_scp_zd, 1)
            vel_ca_squared = numpy.sum(velocity * velocity)
            # polynomial representing range as a function of range distance from SCP
            r_ca = numpy.array([inca.R_CA_SCP, 1], dtype=numpy.float64)

            # the doppler_rate_coeffs represents a polynomial in time, relative to
            #   doppler_rate_ref_time.
            # to construct the doppler centroid polynomial, we need to change scales
            #   to a polynomial in space, relative to SCP.
            doppler_rate_poly = Poly1DType(Coefs=doppler_rate_coeffs)
            alpha = 2.0 / speed_of_light
            t_0 = doppler_rate_ref_time - alpha * inca.R_CA_SCP
            dop_rate_scaled_coeffs = doppler_rate_poly.shift(t_0, alpha, return_poly=False)
            # DRateSFPoly is then a scaled multiple of this scaled poly and r_ca above
            coeffs = -numpy.convolve(dop_rate_scaled_coeffs, r_ca) / (alpha * center_frequency * vel_ca_squared)
            inca.DRateSFPoly = Poly2DType(Coefs=numpy.reshape(coeffs, (coeffs.size, 1)))

            # modify a few of the other fields
            ss_scale = abs(numpy.sqrt(vel_ca_squared)*inca.DRateSFPoly[0, 0])
            grid.Col.SS = abs(col_spacing_zd*ss_scale)
            grid.Col.ImpRespBW = doppler_bandwidth/ss_scale
            inca.TimeCAPoly = Poly1DType(Coefs=[time_scp_zd, -look/ss_scale])

            # doppler centroid
            doppler_cent_poly = Poly1DType(Coefs=doppler_cent_coeffs)
            alpha = 2.0 / speed_of_light
            t_0 = doppler_cent_ref_time - alpha * inca.R_CA_SCP
            scaled_coeffs = doppler_cent_poly.shift(t_0, alpha, return_poly=False)

            # adjust doppler centroid for spotlight, we need to add a second
            # dimension to DopCentroidPoly
            if collection_info.RadarMode.ModeType == 'SPOTLIGHT':
                doppler_cent_est = get_seconds(doppler_cent_time_est, collect_start, precision='us')
                dop_poly = numpy.zeros((scaled_coeffs.shape[0], 2), dtype=numpy.float64)
                dop_poly[0, 1] = -look*center_frequency*alpha*numpy.sqrt(vel_ca_squared)/inca.R_CA_SCP
                dop_poly[1, 1] = -center_frequency*alpha*numpy.sqrt(vel_ca_squared)/(inca.R_CA_SCP ** 2)
                one_way_time = inca.R_CA_SCP / speed_of_light
                pos = position.ARPPoly(doppler_cent_est + one_way_time)
                vel = position.ARPPoly.derivative_eval(doppler_cent_est + one_way_time)
                los = geo_data.SCP.ECF.get_array() - pos
                vel_hat = vel / numpy.linalg.norm(vel)
                dop_poly[:, 0] += -look*(dop_poly[:, 1]*numpy.dot(los, vel_hat))
                inca.DopCentroidPoly = Poly2DType(Coefs=dop_poly)
            else:
                inca.DopCentroidPoly = Poly2DType(Coefs=numpy.reshape(scaled_coeffs, (scaled_coeffs.size, 1)))

            grid.Col.DeltaKCOAPoly = Poly2DType(Coefs=inca.DopCentroidPoly.get_array()*col_spacing_zd/grid.Col.SS)
            # compute grid.Col.DeltaK1/K2 from DeltaKCOAPoly
            coeffs = grid.Col.DeltaKCOAPoly.get_array()[:, 0]
            # get roots
            roots = polynomial.polyroots(coeffs)
            # construct range bounds (in meters)
            range_bounds = (numpy.array([0, image_data.NumRows - 1], dtype=numpy.float64)
                            - image_data.SCPPixel.Row) * grid.Row.SS
            possible_ranges = numpy.copy(range_bounds)
            useful_roots = ((roots > numpy.min(range_bounds)) & (roots < numpy.max(range_bounds)))
            if numpy.any(useful_roots):
                possible_ranges = numpy.concatenate((possible_ranges, roots[useful_roots]), axis=0)
            azimuth_bounds = (numpy.array([0, (image_data.NumCols - 1)], dtype=numpy.float64)
                              - image_data.SCPPixel.Col) * grid.Col.SS
            coords_az_2d, coords_rg_2d = numpy.meshgrid(azimuth_bounds, possible_ranges)
            possible_bounds_deltak = grid.Col.DeltaKCOAPoly(coords_rg_2d, coords_az_2d)
            grid.Col.DeltaK1 = numpy.min(possible_bounds_deltak) - 0.5 * grid.Col.ImpRespBW
            grid.Col.DeltaK2 = numpy.max(possible_bounds_deltak) + 0.5 * grid.Col.ImpRespBW
            # Wrapped spectrum
            if (grid.Col.DeltaK1 < -0.5 / grid.Col.SS) or (grid.Col.DeltaK2 > 0.5 / grid.Col.SS):
                grid.Col.DeltaK1 = -0.5 / abs(grid.Col.SS)
                grid.Col.DeltaK2 = -grid.Col.DeltaK1
            time_coa_poly = fit_time_coa_polynomial(inca, image_data, grid, dop_rate_scaled_coeffs, poly_order=2)
            if collection_info.RadarMode.ModeType == 'SPOTLIGHT':
                # using above was convenience, but not really sensible in spotlight mode
                grid.TimeCOAPoly = Poly2DType(Coefs=[[time_coa_poly.Coefs[0, 0], ], ])
                inca.DopCentroidPoly = None
            elif collection_info.RadarMode.ModeType == 'STRIPMAP':
                # fit TimeCOAPoly for grid
                grid.TimeCOAPoly = time_coa_poly
                inca.DopCentroidCOA = True
            else:
                raise ValueError('unhandled ModeType {}'.format(collection_info.RadarMode.ModeType))
            return RMAType(RMAlgoType='OMEGA_K', INCA=inca)

        def get_radiometric() -> Optional[RadiometricType]:
            def perform_radiometric_fit(component_file: str) -> numpy.ndarray:
                comp_struct = _parse_xml(component_file, without_ns=(self.generation != 'RS2'))
                comp_values = numpy.array(
                    [float(entry) for entry in comp_struct.find('./gains').text.split()], dtype='float64')
                comp_values = 1. / (comp_values * comp_values)  # adjust for sicd convention
                if numpy.all(comp_values == comp_values[0]):
                    return numpy.array([[comp_values[0], ], ], dtype=numpy.float64)
                else:
                    # fit a 1-d polynomial in range
                    if self.generation == 'RS2':
                        coords_rg = \
                            (numpy.arange(image_data.NumRows) - image_data.SCPPixel.Row + image_data.FirstRow)\
                            * grid.Row.SS
                    elif self.generation == 'RCM':  # the rows are sub-sampled
                        start = int(comp_struct.find('./pixelFirstLutValue').text)
                        num_vs = int(comp_struct.find('./numberOfValues').text)
                        t_step = int(comp_struct.find('./stepSize').text)
                        rng_indices = start + numpy.arange(num_vs)*t_step
                        coords_rg = (rng_indices - image_data.SCPPixel.Row + image_data.FirstRow) * grid.Row.SS
                    else:
                        raise ValueError('Unhandled generation {}'.format(self.generation))

                    return numpy.reshape(polynomial.polyfit(coords_rg, comp_values, 3), (-1, 1))

            base_path = os.path.dirname(self.file_name)
            if self.generation == 'RS2':
                beta_file = os.path.join(
                    base_path,
                    self._find(
                        './imageAttributes/lookupTable[@incidenceAngleCorrection="Beta Nought"]').text)
                sigma_file = os.path.join(
                    base_path,
                    self._find(
                        './imageAttributes/lookupTable[@incidenceAngleCorrection="Sigma Nought"]').text)
                gamma_file = os.path.join(
                    base_path,
                    self._find('./imageAttributes/lookupTable[@incidenceAngleCorrection="Gamma"]').text)
            elif self.generation == 'RCM':
                beta_file = os.path.join(
                    base_path, 'calibration',
                    self._find(
                        './imageReferenceAttributes/lookupTableFileName[@sarCalibrationType="Beta Nought"]').text)
                sigma_file = os.path.join(
                    base_path, 'calibration',
                    self._find(
                        './imageReferenceAttributes/lookupTableFileName[@sarCalibrationType="Sigma Nought"]').text)
                gamma_file = os.path.join(
                    base_path, 'calibration',
                    self._find(
                        './imageReferenceAttributes/lookupTableFileName[@sarCalibrationType="Gamma"]').text)
            else:
                raise ValueError(_unhandled_generation_text.format(self.generation))

            if not os.path.isfile(beta_file):
                logger.error(
                    msg="Beta calibration information should be located in file {}, "
                        "which doesn't exist.".format(beta_file))
                return None

            # perform beta, sigma, gamma fit
            beta_zero_sf_poly = perform_radiometric_fit(beta_file)
            sigma_zero_sf_poly = perform_radiometric_fit(sigma_file)
            gamma_zero_sf_poly = perform_radiometric_fit(gamma_file)

            # construct noise poly
            noise_level = None
            if self.generation == 'RS2':
                # noise is in the main product.xml
                beta0_element = self._find('./sourceAttributes/radarParameters'
                                           '/referenceNoiseLevel[@incidenceAngleCorrection="Beta Nought"]')
            elif self.generation == 'RCM':
                noise_file = os.path.join(
                    base_path, 'calibration',
                    self._find('./imageReferenceAttributes/noiseLevelFileName').text)
                noise_root = _parse_xml(noise_file, without_ns=True)
                noise_levels = noise_root.findall('./referenceNoiseLevel')
                beta0s = [entry for entry in noise_levels if entry.find('sarCalibrationType').text.startswith('Beta')]
                beta0_element = beta0s[0] if len(beta0s) > 0 else None
            else:
                raise ValueError(_unhandled_generation_text.format(self.generation))

            if beta0_element is not None:
                pfv = float(beta0_element.find('pixelFirstNoiseValue').text)
                step = float(beta0_element.find('stepSize').text)
                beta0s = numpy.array(
                    [float(x) for x in beta0_element.find('noiseLevelValues').text.split()])
                range_coords = grid.Row.SS * (numpy.arange(len(beta0s)) * step + pfv - image_data.SCPPixel.Row)
                noise_poly = polynomial.polyfit(
                    range_coords,
                    beta0s - 10 * numpy.log10(polynomial.polyval(range_coords, beta_zero_sf_poly[:, 0])), 2)
                noise_level = NoiseLevelType_(
                    NoiseLevelType='ABSOLUTE',
                    NoisePoly=Poly2DType(Coefs=numpy.reshape(noise_poly, (-1, 1))))

            return RadiometricType(BetaZeroSFPoly=beta_zero_sf_poly,
                                   SigmaZeroSFPoly=sigma_zero_sf_poly,
                                   GammaZeroSFPoly=gamma_zero_sf_poly,
                                   NoiseLevel=noise_level)

        def correct_scp() -> None:
            scp_pixel = base_sicd.ImageData.SCPPixel.get_array()
            scp_ecf = base_sicd.project_image_to_ground(scp_pixel, projection_type='HAE')
            base_sicd.update_scp(scp_ecf, coord_system='ECF')

        def get_data_file_names() -> List[str]:
            base_path = os.path.dirname(self.file_name)
            image_files = []
            if self.generation == 'RS2':
                for pol in self._polarizations:
                    fname = self._find('./imageAttributes/fullResolutionImageData[@pole="{}"]'.format(pol))
                    if fname is None:
                        raise ValueError('Got unexpected image file structure.')
                    image_files.append(os.path.join(base_path, fname.text))
            else:
                img_attribute_node = self._find('./sceneAttributes/imageAttributes')
                results = img_attribute_node.findall('./ipdf')

                if len(results) == 1:
                    # there's either a single polarization, or it's one of the NITF files
                    image_files.append(os.path.join(base_path, results[0].text))
                else:
                    for pol in self._polarizations:
                        fname = None
                        for entry in results:
                            if entry.attrib.get('pole', None) == pol:
                                fname = entry.text
                        if fname is None:
                            raise ValueError('Got unexpected image file structure.')
                        image_files.append(os.path.join(base_path, fname))
            return image_files

        center_frequency = float(self._find('./sourceAttributes/radarParameters/radarCenterFrequency').text)
        collect_start = parse_timestring(self._find('./sourceAttributes/rawDataStartTime').text)
        tx_pols, tx_rcv_pols = self._get_sicd_polarizations()

        nitf, collection_info = self._get_sicd_collection_info(collect_start)
        image_creation = self._get_sicd_image_creation()
        position = self._get_sicd_position(collect_start)
        image_data, geo_data = get_image_and_geo_data()
        grid = GridType(ImagePlane='SLANT', Type='RGZERO', Row=get_grid_row(), Col=get_grid_col())
        radar_collection = get_radar_collection()
        timeline = get_timeline()
        image_formation = get_image_formation()
        scpcoa = SCPCOAType(SideOfTrack=self._get_side_of_track())
        rma = get_rma_adjust_grid()
        radiometric = get_radiometric()
        base_sicd = SICDType(
            CollectionInfo=collection_info,
            ImageCreation=image_creation,
            GeoData=geo_data,
            ImageData=image_data,
            Position=position,
            Grid=grid,
            RadarCollection=radar_collection,
            Timeline=timeline,
            ImageFormation=image_formation,
            SCPCOA=scpcoa,
            RMA=rma,
            Radiometric=radiometric,
            _NITF=nitf)
        correct_scp()
        base_sicd.derive()  # derive all the fields
        base_sicd.populate_rniirs(override=False)
        # now, make one copy per polarimetric entry, as appropriate
        the_files = get_data_file_names()

        the_sicds = []
        for i, (original_pol, sicd_pol) in enumerate(zip(self._polarizations, tx_rcv_pols)):
            this_sicd = base_sicd.copy()
            this_sicd.ImageFormation.RcvChanProc.ChanIndices = [i+1, ]
            this_sicd.ImageFormation.TxRcvPolarizationProc = sicd_pol
            the_sicds.append(this_sicd)
        return the_sicds, the_files

    def _get_scansar_sicd(self, beam: str, burst: str) -> Tuple[List[SICDType], List[str]]:
        """
        Gets the SICD collection for the given burst. This is only applicable
        to ScanSAR collects. This will return one SICD per polarimetric collection.
        It will also return the data file(s) for the given beam/burst.

        Parameters
        ----------
        beam : str
        burst : str

        Returns
        -------
        sicds: List[SICDType]
        files: List[str]
        """

        def get_image_and_geo_data() -> Tuple[ImageDataType, GeoDataType, int]:
            img_attributes = self._find('./sceneAttributes/imageAttributes[@burst="{}"]'.format(burst))
            sample_offset = int(img_attributes.find('./pixelOffset').text)
            line_offset = int(img_attributes.find('./lineOffset').text)
            cols = int(img_attributes.find('./numLines').text)
            rows = int(img_attributes.find('./samplesPerLine').text)
            bits_per_sample = self._find('./imageReferenceAttributes/rasterAttributes/bitsPerSample').text
            if bits_per_sample == '32':
                pixel_type = 'RE32F_IM32F'
            elif bits_per_sample == '16':
                pixel_type = 'RE16I_IM16I'
            else:
                raise ValueError('Got unhandled bites per sample {}'.format(bits_per_sample))
            scp_rows = int(0.5*rows)
            scp_cols = int(0.5*cols)
            scp_ecf = self._get_image_location(scp_cols+line_offset, scp_rows+sample_offset)
            im_data = ImageDataType(
                NumRows=rows, NumCols=cols, FirstRow=0, FirstCol=0, PixelType=pixel_type,
                FullImage=(rows, cols), SCPPixel=(scp_rows, scp_cols))
            t_geo_data = GeoDataType(SCP=SCPType(ECF=scp_ecf))
            return im_data, t_geo_data, sample_offset

        def get_grid_row() -> DirParamType:
            row_ss = float(self._find('./imageReferenceAttributes/rasterAttributes/sampledPixelSpacing').text)
            row_irbw = 2*float(self._find('./sourceAttributes'
                                          '/radarParameters'
                                          '/pulseBandwidth[@beam="{}"]'.format(beam)).text)/speed_of_light
            row_wgt_type = WgtTypeType(
                WindowName=self._find('./imageGenerationParameters'
                                      '/sarProcessingInformation'
                                      '/rangeWindow'
                                      '/windowName').text.upper())
            if row_wgt_type.WindowName == 'KAISER':
                row_wgt_type.Parameters = {
                    'BETA': self._find('./imageGenerationParameters'
                                       '/sarProcessingInformation'
                                       '/rangeWindow'
                                       '/windowCoefficient').text}
            return DirParamType(
                SS=row_ss, ImpRespBW=row_irbw, Sgn=-1, KCtr=2*center_frequency/speed_of_light,
                DeltaKCOAPoly=Poly2DType(Coefs=((0,),)), WgtType=row_wgt_type)

        def get_grid_col() -> DirParamType:
            az_win = self._find(
                './imageGenerationParameters/sarProcessingInformation/azimuthWindow[@beam="{}"]'.format(beam))
            col_wgt_type = WgtTypeType(WindowName=az_win.find('./windowName').text.upper())
            if col_wgt_type.WindowName == 'KAISER':
                col_wgt_type.Parameters = {'BETA': az_win.find('./windowCoefficient').text}

            return DirParamType(Sgn=-1, KCtr=0, WgtType=col_wgt_type)

        def get_radar_collection() -> RadarCollectionType:
            radar_params = self._find('./sourceAttributes/radarParameters')
            # Ultrafine and spotlight modes have t pulses, otherwise just one.
            bandwidth_elements = sorted(
                radar_params.findall('pulseBandwidth[@beam="{}"]'.format(beam)), key=lambda x: x.get('pulse'))
            pulse_length_elements = sorted(
                radar_params.findall('pulseLength[@beam="{}"]'.format(beam)), key=lambda x: x.get('pulse'))
            adc_elements = sorted(
                radar_params.findall('adcSamplingRate[@beam="{}"]'.format(beam)), key=lambda x: x.get('pulse'))
            samples_per_echo = float(
                radar_params.find('samplesPerEchoLine[@beam="{}"]'.format(beam)).text)
            wf_params = []
            bandwidths = numpy.empty((len(bandwidth_elements),), dtype=numpy.float64)
            for j, (bwe, ple, adce) in enumerate(zip(bandwidth_elements, pulse_length_elements, adc_elements)):
                bandwidths[j] = float(bwe.text)
                samp_rate = float(adce.text)
                wf_params.append(WaveformParametersType(index=j,
                                                        TxRFBandwidth=float(bwe.text),
                                                        TxPulseLength=float(ple.text),
                                                        ADCSampleRate=samp_rate,
                                                        RcvWindowLength=samples_per_echo / samp_rate,
                                                        RcvDemodType='CHIRP',
                                                        RcvFMRate=0))
            tot_bw = numpy.sum(bandwidths)
            tx_freq = (center_frequency-0.5*tot_bw, center_frequency+0.5*tot_bw)
            t_radar_collection = RadarCollectionType(TxFrequency=tx_freq, Waveform=wf_params)
            t_radar_collection.Waveform[0].TxFreqStart = tx_freq[0]
            for j in range(1, len(bandwidth_elements)):
                t_radar_collection.Waveform[j].TxFreqStart = t_radar_collection.Waveform[j - 1].TxFreqStart + \
                                                             t_radar_collection.Waveform[j - 1].TxRFBandwidth
            t_radar_collection.RcvChannels = [
                ChanParametersType(TxRcvPolarization=entry, index=j+1) for j, entry in enumerate(tx_rcv_pols)]
            if len(tx_pols) == 1:
                t_radar_collection.TxPolarization = tx_pols[0]
            else:
                t_radar_collection.TxPolarization = 'SEQUENCE'
                t_radar_collection.TxSequence = [
                    TxStepType(TxPolarization=entry, index=j+1) for j, entry in enumerate(tx_pols)]
            return t_radar_collection

        def get_timeline():
            # type: () -> TimelineType
            ipp = IPPSetType(
                index=0, TStart=0, TEnd=processing_time_span,
                IPPStart=0, IPPEnd=num_of_pulses - 1,
                IPPPoly=Poly1DType(Coefs=(0, pulse_rep_freq)))

            return TimelineType(
                CollectStart=collect_start,
                CollectDuration=processing_time_span,
                IPP=[ipp, ])

        def get_image_formation() -> ImageFormationType:
            pulse_parts = len(
                self._findall('./sourceAttributes/radarParameters/pulseBandwidth[@beam="{}"]'.format(beam)))
            return ImageFormationType(
                # PRFScaleFactor for either polarimetric or multi-step, but not both.
                RcvChanProc=RcvChanProcType(
                    NumChanProc=1, PRFScaleFactor=1./max(pulse_parts, len(tx_pols))),
                ImageFormAlgo='RMA',
                TStartProc=timeline.IPP[0].TStart,
                TEndProc=timeline.IPP[0].TEnd,
                TxFrequencyProc=(
                    radar_collection.TxFrequency.Min, radar_collection.TxFrequency.Max),
                STBeamComp='GLOBAL',
                ImageBeamComp='SV',
                AzAutofocus='NO',
                RgAutofocus='NO')

        def get_rma_adjust_grid() -> RMAType:
            sar_processing_info = self._find('./imageGenerationParameters/sarProcessingInformation')

            doppler_bandwidth = float(sar_processing_info.find('./totalProcessedAzimuthBandwidth').text)
            zero_dop_last_line = parse_timestring(
                sar_processing_info.find('./zeroDopplerTimeLastLine').text, precision='us')
            zero_dop_first_line = parse_timestring(
                sar_processing_info.find('./zeroDopplerTimeFirstLine').text, precision='us')

            # doppler rate coefficients
            doppler_rate_coeffs = numpy.array(
                [float(entry) for entry in dop_rate_estimate_node.find('./dopplerRateCoefficients').text.split()],
                dtype='float64')
            doppler_rate_ref_time = float(dop_rate_estimate_node.find('./dopplerRateReferenceTime').text)

            col_spacing_zd = \
                get_seconds(zero_dop_last_line, zero_dop_first_line, precision='us')/self._num_lines_processed
            # NB: this will be negative for Ascending

            # zero doppler time of SCP relative to collect start
            time_scp_zd = 0.5*processing_time_span

            r_ca_scp = numpy.linalg.norm(position.ARPPoly(time_scp_zd) - geo_data.SCP.ECF.get_array())
            inca = INCAType(R_CA_SCP=r_ca_scp, FreqZero=center_frequency)
            # doppler rate calculations
            velocity = position.ARPPoly.derivative_eval(time_scp_zd, der_order=1)
            vel_ca_squared = numpy.sum(velocity * velocity)
            # polynomial representing range as a function of range distance from SCP
            r_ca = numpy.array([inca.R_CA_SCP, 1], dtype='float64')

            # the doppler_rate_coeffs represents a polynomial in time, relative to
            #   doppler_rate_ref_time.
            # to construct the doppler centroid polynomial, we need to change scales
            #   to a polynomial in space, relative to SCP.
            doppler_rate_poly = Poly1DType(Coefs=doppler_rate_coeffs)
            alpha = 2.0 / speed_of_light
            t_0 = doppler_rate_ref_time - alpha*inca.R_CA_SCP
            dop_rate_scaled_coeffs = doppler_rate_poly.shift(t_0, alpha, return_poly=False)
            # DRateSFPoly is then a scaled multiple of this scaled poly and r_ca above
            coeffs = -numpy.convolve(dop_rate_scaled_coeffs, r_ca) / (alpha * center_frequency * vel_ca_squared)
            inca.DRateSFPoly = Poly2DType(Coefs=numpy.reshape(coeffs, (coeffs.size, 1)))

            look = scpcoa.look

            # modify a few of the other fields
            ss_scale = abs(numpy.sqrt(vel_ca_squared)*inca.DRateSFPoly[0, 0])
            grid.Col.SS = abs(col_spacing_zd*ss_scale)
            grid.Col.ImpRespBW = doppler_bandwidth/ss_scale
            inca.TimeCAPoly = Poly1DType(Coefs=[time_scp_zd, -look/ss_scale])

            # doppler centroid
            doppler_cent_coeffs = numpy.array(
                [float(entry) for entry in
                 dop_centroid_estimate_node.find('./dopplerCentroidCoefficients').text.split()],
                dtype='float64')
            doppler_cent_ref_time = float(
                dop_centroid_estimate_node.find('./dopplerCentroidReferenceTime').text)

            doppler_cent_poly = Poly1DType(Coefs=doppler_cent_coeffs)
            alpha = 2.0 / speed_of_light
            t_0 = doppler_cent_ref_time - alpha * inca.R_CA_SCP
            scaled_coeffs = doppler_cent_poly.shift(t_0, alpha, return_poly=False)

            # adjust doppler centroid for spotlight, we need to add a second
            # dimension to DopCentroidPoly
            if collection_info.RadarMode.ModeType == 'SPOTLIGHT':
                doppler_cent_est = get_seconds(dop_centroid_est_time, collect_start, precision='us')
                dop_poly = numpy.zeros((scaled_coeffs.shape[0], 2), dtype=numpy.float64)
                dop_poly[0, 1] = -look*center_frequency*alpha*numpy.sqrt(vel_ca_squared)/inca.R_CA_SCP
                dop_poly[1, 1] = -center_frequency*alpha*numpy.sqrt(vel_ca_squared)/(inca.R_CA_SCP**2)
                one_way_time = inca.R_CA_SCP / speed_of_light
                use_time = doppler_cent_est + one_way_time
                pos = position.ARPPoly(use_time)
                vel = position.ARPPoly.derivative_eval(use_time, der_order=1)
                los = geo_data.SCP.ECF.get_array() - pos
                vel_hat = vel / numpy.linalg.norm(vel)
                dop_poly[:, 0] += -look*(dop_poly[:, 1]*numpy.dot(los, vel_hat))
                inca.DopCentroidPoly = Poly2DType(Coefs=dop_poly)  # NB: this is set for use in fit-time_coa_poly
            else:
                raise ValueError('ScanSAR mode data should be SPOTLIGHT mode')

            grid.Col.DeltaKCOAPoly = Poly2DType(Coefs=inca.DopCentroidPoly.get_array()*col_spacing_zd/grid.Col.SS)
            # compute grid.Col.DeltaK1/K2 from DeltaKCOAPoly
            coeffs = grid.Col.DeltaKCOAPoly.get_array()[:, 0]
            # get roots
            roots = polynomial.polyroots(coeffs)
            # construct range bounds (in meters)
            range_bounds = (numpy.array([0, image_data.NumRows - 1], dtype='float64')
                            - image_data.SCPPixel.Row) * grid.Row.SS
            possible_ranges = numpy.copy(range_bounds)
            useful_roots = ((roots > numpy.min(range_bounds)) & (roots < numpy.max(range_bounds)))
            if numpy.any(useful_roots):
                possible_ranges = numpy.concatenate((possible_ranges, roots[useful_roots]), axis=0)
            azimuth_bounds = (numpy.array([0, (image_data.NumCols - 1)], dtype='float64')
                              - image_data.SCPPixel.Col) * grid.Col.SS
            coords_az_2d, coords_rg_2d = numpy.meshgrid(azimuth_bounds, possible_ranges)
            possible_bounds_deltak = grid.Col.DeltaKCOAPoly(coords_rg_2d, coords_az_2d)
            grid.Col.DeltaK1 = numpy.min(possible_bounds_deltak) - 0.5 * grid.Col.ImpRespBW
            grid.Col.DeltaK2 = numpy.max(possible_bounds_deltak) + 0.5 * grid.Col.ImpRespBW
            # Wrapped spectrum
            if (grid.Col.DeltaK1 < -0.5 / grid.Col.SS) or (grid.Col.DeltaK2 > 0.5 / grid.Col.SS):
                grid.Col.DeltaK1 = -0.5 / abs(grid.Col.SS)
                grid.Col.DeltaK2 = -grid.Col.DeltaK1
            time_coa_poly = fit_time_coa_polynomial(inca, image_data, grid, dop_rate_scaled_coeffs, poly_order=2)
            if collection_info.RadarMode.ModeType == 'SPOTLIGHT':
                # using above was convenience, but not really sensible in spotlight mode
                grid.TimeCOAPoly = Poly2DType(Coefs=[[time_coa_poly.Coefs[0, 0], ], ])
                inca.DopCentroidPoly = None
            else:
                raise ValueError('ScanSAR mode data should be SPOTLIGHT mode')
            return RMAType(RMAlgoType='OMEGA_K', INCA=inca)

        def get_radiometric() -> Optional[RadiometricType]:
            def perform_radiometric_fit(component_file: str) -> numpy.ndarray:
                comp_struct = _parse_xml(component_file, without_ns=True)
                comp_values = numpy.array(
                    [float(entry) for entry in comp_struct.find('./gains').text.split()], dtype=numpy.float64)
                comp_values = 1. / (comp_values * comp_values)  # adjust for sicd convention
                if numpy.all(comp_values == comp_values[0]):
                    return numpy.array([[comp_values[0], ], ], dtype=numpy.float64)
                else:
                    # fit a 1-d polynomial in range
                    t_pfv = int(comp_struct.find('./pixelFirstLutValue').text)
                    t_step = int(comp_struct.find('./stepSize').text)
                    num_vs = int(comp_struct.find('./numberOfValues').text)
                    t_range_inds = t_pfv + numpy.arange(num_vs)*t_step
                    coords_rg = grid.Row.SS * (t_range_inds - (image_data.SCPPixel.Row + row_shift))
                    return numpy.reshape(polynomial.polyfit(coords_rg, comp_values, 3), (-1, 1))

            # NB: I am neglecting the difference in polarization for these
            base_path = os.path.dirname(self.file_name)
            beta_file = os.path.join(
                base_path, 'calibration',
                self._find('./imageReferenceAttributes/lookupTableFileName[@sarCalibrationType="Beta Nought"]').text)
            sigma_file = os.path.join(
                base_path, 'calibration',
                self._find('./imageReferenceAttributes/lookupTableFileName[@sarCalibrationType="Sigma Nought"]').text)
            gamma_file = os.path.join(
                base_path, 'calibration',
                self._find('./imageReferenceAttributes/lookupTableFileName[@sarCalibrationType="Gamma"]').text)
            if not os.path.isfile(beta_file):
                logger.error(
                    msg="Beta calibration information should be located in file {}, "
                        "which doesn't exist.".format(beta_file))
                return None

            # perform beta, sigma, gamma fit
            beta_zero_sf_poly = perform_radiometric_fit(beta_file)
            sigma_zero_sf_poly = perform_radiometric_fit(sigma_file)
            gamma_zero_sf_poly = perform_radiometric_fit(gamma_file)

            # construct noise poly, neglecting the polarization again
            noise_level = None
            noise_stem = self._find('./imageReferenceAttributes/noiseLevelFileName').text
            noise_file = os.path.join(base_path, 'calibration', noise_stem)
            noise_root = _parse_xml(noise_file, without_ns=True)

            noise_levels = noise_root.findall('./perBeamReferenceNoiseLevel')
            beta0s = [entry for entry in noise_levels if (entry.find('./sarCalibrationType').text.startswith('Beta')
                                                          and entry.find('./beam').text == beam)]
            beta0_element = beta0s[0] if len(beta0s) > 0 else None

            if beta0_element is not None:
                pfv = float(beta0_element.find('pixelFirstNoiseValue').text)
                step = float(beta0_element.find('stepSize').text)
                beta0s = numpy.array(
                    [float(x) for x in beta0_element.find('noiseLevelValues').text.split()])
                range_inds = pfv + numpy.arange(len(beta0s)) * step
                range_coords = grid.Row.SS * (range_inds - (image_data.SCPPixel.Row + row_shift))
                noise_value = beta0s - 10*numpy.log10(polynomial.polyval(range_coords, beta_zero_sf_poly[:, 0]))
                noise_poly = polynomial.polyfit(range_coords, noise_value, 2)
                noise_level = NoiseLevelType_(
                    NoiseLevelType='ABSOLUTE',
                    NoisePoly=Poly2DType(Coefs=numpy.reshape(noise_poly, (-1, 1))))

            return RadiometricType(BetaZeroSFPoly=beta_zero_sf_poly,
                                   SigmaZeroSFPoly=sigma_zero_sf_poly,
                                   GammaZeroSFPoly=gamma_zero_sf_poly,
                                   NoiseLevel=noise_level)

        def correct_scp() -> None:
            scp_pixel = base_sicd.ImageData.SCPPixel.get_array()
            scp_ecf = base_sicd.project_image_to_ground(scp_pixel, projection_type='HAE')
            base_sicd.GeoData.SCP.ECF = scp_ecf

        def get_data_file_names() -> List[str]:
            base_path = os.path.dirname(self.file_name)
            image_files = []
            img_attribute_node = self._find('./sceneAttributes/imageAttributes[@burst="{}"]'.format(burst))
            results = img_attribute_node.findall('./ipdf')

            if len(results) == 1:
                # there's either a single polarization, or it's one of the NITF files
                image_files.append(os.path.join(base_path, results[0].text))
            else:
                for pol in self._polarizations:
                    fname = None
                    for entry in results:
                        if entry.attrib.get('pole', None) == pol:
                            fname = entry.text
                    if fname is None:
                        raise ValueError('Got unexpected image file structure for burst {}'.format(burst))
                    image_files.append(os.path.join(base_path, fname))
            return image_files

        if self.generation != 'RCM':
            raise ValueError('Unhandled generation {}'.format(self.generation))

        center_frequency = float(self._find('./sourceAttributes/radarParameters/radarCenterFrequency').text)
        tx_pols, tx_rcv_pols = self._get_sicd_polarizations()

        # extract some common use doppler information
        num_of_pulses = int(
            self._find(
                './sourceAttributes/radarParameters/numberOfPulseIntervalsPerDwell[@beam="{}"]'.format(beam)).text)
        # NB: I am neglecting that prfInformation is provided separately for pols data because
        #   it appears identical
        pulse_rep_freq = float(
            self._find('./sourceAttributes'
                       '/radarParameters'
                       '/prfInformation[@beam="{}"]'.format(beam) +
                       '/pulseRepetitionFrequency').text)
        dop_rate_estimate_node = self._find(
            './dopplerRate/dopplerRateEstimate[@burst="{}"]'.format(burst))
        dop_centroid_estimate_node = self._find(
            './dopplerCentroid/dopplerCentroidEstimate[@burst="{}"]'.format(burst))
        dop_centroid_est_time = parse_timestring(
            dop_centroid_estimate_node.find(
                './timeOfDopplerCentroidEstimate').text, precision='us')
        processing_time_span = (num_of_pulses - 1)/pulse_rep_freq  # in seconds
        collect_start = dop_centroid_est_time - int(0.5*(processing_time_span*1000000))

        nitf, collection_info = self._get_sicd_collection_info(collect_start)
        image_creation = self._get_sicd_image_creation()
        position = self._get_sicd_position(collect_start)
        image_data, geo_data, row_shift = get_image_and_geo_data()
        grid = GridType(ImagePlane='SLANT', Type='RGZERO', Row=get_grid_row(), Col=get_grid_col())
        radar_collection = get_radar_collection()
        timeline = get_timeline()
        image_formation = get_image_formation()
        scpcoa = SCPCOAType(SideOfTrack=self._get_side_of_track())
        rma = get_rma_adjust_grid()
        radiometric = get_radiometric()
        base_sicd = SICDType(
            CollectionInfo=collection_info,
            ImageCreation=image_creation,
            GeoData=geo_data,
            ImageData=image_data,
            Position=position,
            Grid=grid,
            RadarCollection=radar_collection,
            Timeline=timeline,
            ImageFormation=image_formation,
            SCPCOA=scpcoa,
            RMA=rma,
            Radiometric=radiometric,
            _NITF=nitf)
        correct_scp()
        base_sicd.derive()  # derive all the fields
        base_sicd.populate_rniirs(override=False)
        # now, make one copy per polarimetric entry, as appropriate
        the_files = get_data_file_names()

        the_sicds = []
        for i, (original_pol, sicd_pol) in enumerate(zip(self._polarizations, tx_rcv_pols)):
            this_sicd = base_sicd.copy()
            this_sicd.ImageFormation.RcvChanProc.ChanIndices = [i+1, ]
            this_sicd.ImageFormation.TxRcvPolarizationProc = sicd_pol
            the_sicds.append(this_sicd)
        return the_sicds, the_files

    def get_sicd_collection(self) -> Tuple[List[List[SICDType]], List[List[str]]]:
        """
        Gets the collection of sicd objects.

        Returns
        -------
        sicds: List[List[SICDType]]
        files: List[List[str]]
        """

        sicds = []
        data_files = []
        if self._bursts is None:
            t_sicds, t_files = self._get_regular_sicd()
            sicds.append(t_sicds)
            data_files.append(t_files)
        else:
            for (beam, burst) in self._bursts:
                t_sicds, t_files = self._get_scansar_sicd(beam, burst)
                sicds.append(t_sicds)
                data_files.append(t_files)
        return sicds, data_files


##############
# reader implementation - really just borrows from tiff or NITF reader

class RadarSatReader(SICDTypeReader):
    """
    A RadarSat-2 and RadarSat Constellation Mission (RCM) SLC file package
    reader implementation.

    **Changed in version 1.3.0** for reading changes.
    """

    __slots__ = ('_radar_sat_details', '_other_reader')

    def __init__(self, radar_sat_details):
        """

        Parameters
        ----------
        radar_sat_details : str|RadarSatDetails
            file name or RadarSatDetails object
        """

        self._other_reader = None
        if isinstance(radar_sat_details, str):
            radar_sat_details = RadarSatDetails(radar_sat_details)
        if not isinstance(radar_sat_details, RadarSatDetails):
            raise TypeError('The input argument for RadarSatReader must be a '
                            'filename or RadarSatDetails object')
        self._radar_sat_details = radar_sat_details
        # determine symmetry
        reverse_axes, transpose_axes = self._radar_sat_details.get_symmetry()
        # get the sicd collection and data file names
        the_sicds, the_files = self.radarsat_details.get_sicd_collection()
        use_sicds = []
        the_segments = []
        for sicd_entry, file_entry in zip(the_sicds, the_files):
            the_segments.extend(self._construct_segments(sicd_entry, file_entry, reverse_axes, transpose_axes))
            use_sicds.extend(sicd_entry)

        SICDTypeReader.__init__(self, the_segments, use_sicds, close_segments=True)
        self._check_sizes()

    def _construct_segments(
            self,
            sicds: List[SICDType],
            data_files: List[str],
            reverse_axes: Optional[Tuple[int, ...]],
            transpose_axes: Optional[Tuple[int, ...]]) -> List[DataSegment]:
        """
        Construct the data segments.

        Parameters
        ----------
        sicds : List[SICDType]
        data_files : List[str]
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]

        Returns
        -------
        List[BaseChipper]
        """

        data_segments = []
        if len(sicds) == len(data_files):
            for sicd, data_file in zip(sicds, data_files):
                fext = os.path.splitext(data_file)[1]
                if fext in ['.tiff', '.tif']:
                    data_segments.append(_construct_tiff_segment(sicd, data_file, reverse_axes, transpose_axes))
                elif fext in ['.nitf', '.ntf']:
                    reader, segment = _construct_single_nitf_segment(sicd, data_file, reverse_axes, transpose_axes)
                    self._other_reader = reader  # NB: maintain reference, to keep segment open
                    data_segments.append(segment)
                else:
                    raise ValueError(
                        'The radarsat reader requires image files in tiff or nitf format. '
                        'Uncertain how to interpret file {}'.format(data_file))
        elif len(data_files) == 1:
            data_file = data_files[0]
            fext = os.path.splitext(data_file)[1]
            if fext not in ['.nitf', '.ntf']:
                raise ValueError(
                    'The radarsat has image data for multiple polarizations provided in a '
                    'single image file. This requires an image files in nitf format. '
                    'Uncertain how to interpret file {}'.format(data_file))
            reader, segments = _construct_multiple_nitf_segment(sicds, data_file, reverse_axes, transpose_axes)
            self._other_reader = reader  # NB: maintain reference, to keep segments open
            data_segments.extend(segments)
        else:
            raise ValueError(
                'Unclear how to construct chipper elements for {} sicd elements '
                'from {} image files.'.format(len(sicds), len(data_files)))
        return data_segments

    @property
    def radarsat_details(self) -> RadarSatDetails:
        """
        RadarSarDetails: The RadarSat/RCM details object.
        """

        return self._radar_sat_details

    @property
    def file_name(self) -> str:
        return self.radarsat_details.directory_name

    def close(self) -> None:
        SICDTypeReader.close(self)
        self._other_reader = None


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name: str) -> Optional[RadarSatReader]:
    """
    Tests whether a given file_name corresponds to a RadarSat file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    RadarSatReader|None
        `RadarSatReader` instance if RadarSat file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    try:
        details = RadarSatDetails(file_name)
        logger.info('Path {} is determined to be or contain a RadarSat or RCM product.xml file.'.format(file_name))
        return RadarSatReader(details)
    except SarpyIOError:
        return None



"""
Window function definitions and a few helper functions. This just passes through
to scipy functions after managing scipy version dependent import structure.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, Optional

import scipy
import numpy
from scipy.optimize import newton


_version_string_parts = scipy.__version__.split('.')
_version = (int(_version_string_parts[0]), int(_version_string_parts[1]))


############
# some basic window function definitions

if _version >= (1, 1):
    # noinspection PyUnresolvedReferences
    from scipy.signal.windows import general_hamming as _general_hamming, \
        kaiser as _kaiser
else:
    _general_hamming = None
    # noinspection PyUnresolvedReferences
    from scipy.signal import kaiser as _kaiser

if _version >= (1, 6):
    # noinspection PyUnresolvedReferences
    from scipy.signal.windows import taylor as _taylor
else:
    _taylor = None


def general_hamming(
        M: int,
        alpha: float,
        sym: bool = True) -> numpy.ndarray:
    r"""
    Returns a generalized hamming function. Constructed (non-symmetric) as
    :math:`\alpha - (1-\alpha)\cos\left(\frac{2\pi n}{M-1}\right) 0\leq n \leq M-1`

    Parameters
    ----------
    M : int
        Number of points in the output window.
    alpha : float
        The window coefficient.
    sym : bool
        When `True` (default), generates a symmetric window, for use in filter
        design. When `False`, generates a periodic window, for use in spectral analysis.

    Returns
    -------
    numpy.ndarray
    """

    if _general_hamming is not None:
        return _general_hamming(M, alpha, sym=sym)

    if (M % 2) == 0:
        k = int(M / 2)
    else:
        k = int((M + 1) / 2)
    theta = 2 * numpy.pi * numpy.arange(k) / (M - 1)

    weights = numpy.zeros((M,), dtype=numpy.float64)
    if sym:
        weights[:k] = (alpha - (1 - alpha) * numpy.cos(theta))
        weights[k:] = weights[k - 1::-1]
    else:
        weights[:k] = (alpha - (1 - alpha) * numpy.cos(theta))[::-1]
        weights[k:] = weights[k - 1::-1]
    return weights


def hamming(
        M: int,
        sym: bool = True) -> numpy.ndarray:
    """
    The hamming window, which is a general hamming window with alpha=0.54.

    Parameters
    ----------
    M : int
        Number of points in the output window.
    sym : bool
        When `True` (default), generates a symmetric window, for use in filter
        design. When `False`, generates a periodic window, for use in spectral analysis.

    Returns
    -------
    numpy.ndarray
    """

    return general_hamming(M, 0.54, sym=sym)


def hanning(
        M: int,
        sym: bool = True) -> numpy.ndarray:
    """
    The hanning or hann window, which is a general hamming window with alpha=0.5.

    Parameters
    ----------
    M : int
        Number of points in the output window.
    sym : bool
        When `True` (default), generates a symmetric window, for use in filter
        design. When `False`, generates a periodic window, for use in spectral analysis.

    Returns
    -------
    numpy.ndarray
    """

    return general_hamming(M, 0.5, sym=sym)


def taylor(
        M: int,
        nbar: int = 4,
        sll: float = -30,
        norm: bool = True,
        sym: bool = True) -> numpy.ndarray:
    """
    The Taylor window taper function approximates the Dolph-Chebyshev windows
    constant sidelobe level for a parameterized number of near-in sidelobes,
    but then allows a taper beyond.

    The SAR (synthetic aperature radar) community commonly uses Taylor weighting
    for image formation processing because it provides strong, selectable sidelobe
    suppression with minimum broadening of the mainlobe.

    Parameters
    ----------
    M : int
        Number of points in the output window.
    nbar : int
        Number of nearly constant level sidelobes adjacent to the mainlobe.
    sll : float
        Desired suppression of sidelobe level in decibels (dB) relative to the
        DC gain of the mainlobe. This should be a positive number.
    norm : bool
        When `True` (default), divides the window by the largest (middle) value
        for odd-length windows or the value that would occur between the two
        repeated middle values for even-length windows such that all values are
        less than or equal to 1. When `False` the DC gain will remain at 1 (0 dB)
        and the sidelobes will be sll dB down.
    sym : bool
        When `True` (default), generates a symmetric window, for use in filter
        design. When `False`, generates a periodic window, for use in spectral analysis.

    Returns
    -------
    numpy.ndarray
    """

    if _taylor is not None:
        if sll < 0:
            sll *= -1
        return _taylor(M, nbar=nbar, sll=sll, norm=norm, sym=sym)

    if sll > 0:
        sll *= -1

    a = numpy.arccosh(10**(-sll/20.))/numpy.pi
    # Taylor pulse widening (dilation) factor
    sp2 = (nbar*nbar)/(a*a + (nbar-0.5)*(nbar-0.5))
    # the angular space in n points
    xi = numpy.linspace(-numpy.pi, numpy.pi, M)
    # calculate the cosine weights
    out = numpy.ones((M, ), dtype=numpy.float64)  # the "constant" term
    max_value = 1.0

    coefs = numpy.arange(1, nbar)
    sgn = 1

    for m in coefs:
        coefs1 = (coefs - 0.5)
        coefs2 = coefs[coefs != m]
        numerator = numpy.prod(1 - (m*m)/(sp2*(a*a + coefs1*coefs1)))
        denominator = numpy.prod(1 - (m*m)/(coefs2*coefs2))
        out += sgn*(numerator/denominator)*numpy.cos(m*xi)
        max_value += sgn*(numerator/denominator)
        sgn *= -1

    if not sym:
        k = int(M/2)
        l = M-k
        out2 = numpy.empty((M, ), dtype='float64')
        out2[:k] = out[l:]
        out2[k:] = out[:l]
        out = out2

    if norm:
        out /= max_value
    return out


def kaiser(
        M: int,
        beta: float,
        sym: bool = True) -> numpy.ndarray:
    """
    Return a Kaiser window, which is a taper formed by using a Bessel function.

    Parameters
    ----------
    M : int
        Number of points in the output window.
    beta : float
        Shape parameter, determines trade-off between main-lobe width and side
        lobe level. As beta gets large, the window narrows.
    sym : bool
        When `True` (default), generates a symmetric window, for use in filter
        design. When `False`, generates a periodic window, for use in spectral analysis.

    Returns
    -------
    numpy.ndarray
    """

    return _kaiser(M, beta, sym=sym)


#################
# helper methods

def hamming_ipr(
        x: Union[numpy.ndarray, float],
        a: float) -> Union[numpy.ndarray, float]:
    """
    Evaluate the Hamming impulse response function over the given array.

    Parameters
    ----------
    x : numpy.ndarray|float|int
    a : float
        The Hamming parameter value.

    Returns
    -------
    numpy.ndarray
    """

    return a*numpy.sinc(x) + 0.5*(1-a)*(numpy.sinc(x-1) + numpy.sinc(x+1)) - a/numpy.sqrt(2)


def get_hamming_broadening_factor(coef: float) -> float:
    test_array = numpy.linspace(0.3, 2.5, 100)
    values = hamming_ipr(test_array, coef)
    init_value = test_array[numpy.argmin(numpy.abs(values))]
    zero = newton(hamming_ipr, init_value, args=(coef,), tol=1e-12, maxiter=100)
    return 2 * zero


def find_half_power(
        wgt_funct: Optional[numpy.ndarray],
        oversample: int = 1024) -> Optional[float]:
    """
    Find the half power point of the impulse response function.

    Parameters
    ----------
    wgt_funct : None|numpy.ndarray
    oversample : int

    Returns
    -------
    None|float
    """

    if wgt_funct is None:
        return None

    # solve for the half-power point in an oversampled impulse response
    impulse_response = numpy.abs(numpy.fft.fft(wgt_funct, wgt_funct.size*oversample))/numpy.sum(wgt_funct)
    ind = numpy.flatnonzero(impulse_response < 1 / numpy.sqrt(2))[0]
    # find first index with less than half power,
    # then linearly interpolate to estimate 1/sqrt(2) crossing
    v0 = impulse_response[ind - 1]
    v1 = impulse_response[ind]
    zero_ind = ind - 1 + (1./numpy.sqrt(2) - v0)/(v1 - v0)
    return 2*zero_ind/oversample




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class IMASDAType(TREElement):
    def __init__(self, value):
        super(IMASDAType, self).__init__()
        self.add_field('LONTR', 's', 22, value)
        self.add_field('LATTR', 's', 22, value)
        self.add_field('ELVTR', 's', 22, value)
        self.add_field('LONSC', 's', 22, value)
        self.add_field('LATSC', 's', 22, value)
        self.add_field('ELVSC', 's', 22, value)
        self.add_field('XITR', 's', 22, value)
        self.add_field('YITR', 's', 22, value)
        self.add_field('XISC', 's', 22, value)
        self.add_field('YISC', 's', 22, value)
        self.add_field('DELEV', 's', 22, value)


class IMASDA(TREExtension):
    _tag_value = 'IMASDA'
    _data_type = IMASDAType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class XINC(TREElement):
    def __init__(self, value):
        super(XINC, self).__init__()
        self.add_field('XINC', 's', 22, value)


class XIDC(TREElement):
    def __init__(self, value):
        super(XIDC, self).__init__()
        self.add_field('XIDC', 's', 22, value)


class YINC(TREElement):
    def __init__(self, value):
        super(YINC, self).__init__()
        self.add_field('YINC', 's', 22, value)


class YIDC(TREElement):
    def __init__(self, value):
        super(YIDC, self).__init__()
        self.add_field('YIDC', 's', 22, value)


class IMRFCAType(TREElement):
    def __init__(self, value):
        super(IMRFCAType, self).__init__()
        self.add_loop('XINCs', 20, XINC, value)
        self.add_loop('XIDCs', 20, XIDC, value)
        self.add_loop('YINCs', 20, YINC, value)
        self.add_loop('YIDCs', 20, YIDC, value)


class IMRFCA(TREExtension):
    _tag_value = 'IMRFCA'
    _data_type = IMRFCAType



"""
Common functionality for converting metadata
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Iterator, Tuple, List, Optional, Union

import numpy
from numpy.polynomial import polynomial
from scipy.stats import scoreatpercentile
from scipy.linalg import lstsq

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.blocks import Poly2DType

logger = logging.getLogger(__name__)


def two_dim_poly_fit(
        x: numpy.ndarray,
        y: numpy.ndarray,
        z: numpy.ndarray,
        x_order: int = 2,
        y_order: int = 2,
        x_scale: float = 1.,
        y_scale: float = 1.,
        rcond: Optional[float] = None) -> Tuple[
            numpy.ndarray, numpy.ndarray, int, numpy.ndarray]:
    """
    Perform fit of data to two-dimensional polynomial.

    Parameters
    ----------
    x : numpy.ndarray
        the x data
    y : numpy.ndarray
        the y data
    z : numpy.ndarray
        the z data
    x_order : int
        the order for x
    y_order : int
        the order for y
    x_scale : float
        In order to help the fitting problem to become better conditioned, the independent
        variables can be scaled, the fit performed, and then the solution rescaled.
    y_scale : float
    rcond : None|float
        passed through to :func:`scipy.linalg.lstsq` (as `cond`).

    Returns
    -------
    solution : numpy.ndarray
        the coefficient array
    residuals : numpy.ndarray
        this is often an empty array
    rank : int
        the rank of the fitting matrix
    sing_values : numpy.ndarray
        the singular values of the fitting matrix
    """

    if not isinstance(x, numpy.ndarray) or not isinstance(y, numpy.ndarray) or not isinstance(z, numpy.ndarray):
        raise TypeError('x, y, z must be numpy arrays')
    if (x.size != z.size) or (y.size != z.size):
        raise ValueError('x, y, z must have the same cardinality size.')

    x = x.flatten()*x_scale
    y = y.flatten()*y_scale
    z = z.flatten()
    # first, we need to formulate this as A*t = z
    # where A has shape (x.size, (x_order+1)*(y_order+1))
    # and t has shape ((x_order+1)*(y_order+1), )
    A = numpy.empty((x.size, (x_order+1)*(y_order+1)), dtype=numpy.float64)
    # noinspection PyTypeChecker
    for i, index in enumerate(numpy.ndindex((x_order+1, y_order+1))):
        A[:, i] = numpy.power(x, index[0])*numpy.power(y, index[1])
    # perform least squares fit
    sol, residuals, rank, sing_values = lstsq(A, z, cond=rcond)
    if isinstance(residuals, (numpy.ndarray, numpy.number)):
        residuals /= float(x.size)
    sol = numpy.power(x_scale, numpy.arange(x_order+1))[:, numpy.newaxis] * \
        numpy.reshape(sol, (x_order+1, y_order+1)) * \
        numpy.power(y_scale, numpy.arange(y_order+1))
    return sol, residuals, rank, sing_values


def get_im_physical_coords(
        array,
        grid,
        image_data,
        direction):
    """
    Converts one dimension of "pixel" image (row or column) coordinates to
    "physical" image (range or azimuth in meters) coordinates, for use in the
    various two-variable sicd polynomials.

    Parameters
    ----------
    array : numpy.ndarray|float|int
        either row or col coordinate component
    grid : sarpy.io.complex.sicd_elements.Grid.GridType
    image_data : sarpy.io.complex.sicd_elements.ImageData.ImageDataType
    direction : str
        one of 'Row' or 'Col' (case insensitive) to determine which
    Returns
    -------
    numpy.array|float
    """

    if direction.upper() == 'ROW':
        return (array - image_data.SCPPixel.Row + image_data.FirstRow)*grid.Row.SS
    elif direction.upper() == 'COL':
        return (array - image_data.SCPPixel.Col + image_data.FirstCol)*grid.Col.SS
    else:
        raise ValueError('Unrecognized direction {}'.format(direction))


def fit_time_coa_polynomial(
        inca,
        image_data,
        grid,
        dop_rate_scaled_coeffs,
        poly_order=2):
    """

    Parameters
    ----------
    inca : sarpy.io.complex.sicd_elements.RMA.INCAType
    image_data : sarpy.io.complex.sicd_elements.ImageData.ImageDataType
    grid : sarpy.io.complex.sicd_elements.Grid.GridType
    dop_rate_scaled_coeffs : numpy.ndarray
        the dop rate polynomial relative to physical coordinates - this a
        common construct in converting metadata for csk/sentinel/radarsat
    poly_order : int
        the degree of the polynomial to fit.
    Returns
    -------
    Poly2DType
    """

    grid_samples = poly_order + 8
    coords_az = get_im_physical_coords(
        numpy.linspace(0, image_data.NumCols - 1, grid_samples, dtype='float64'), grid, image_data, 'col')
    coords_rg = get_im_physical_coords(
        numpy.linspace(0, image_data.NumRows - 1, grid_samples, dtype='float64'), grid, image_data, 'row')
    coords_az_2d, coords_rg_2d = numpy.meshgrid(coords_az, coords_rg)
    time_ca_sampled = inca.TimeCAPoly(coords_az_2d)
    dop_centroid_sampled = inca.DopCentroidPoly(coords_rg_2d, coords_az_2d)
    doppler_rate_sampled = polynomial.polyval(coords_rg_2d, dop_rate_scaled_coeffs)
    time_coa_sampled = time_ca_sampled + dop_centroid_sampled / doppler_rate_sampled
    coefs, residuals, rank, sing_values = two_dim_poly_fit(
        coords_rg_2d, coords_az_2d, time_coa_sampled,
        x_order=poly_order, y_order=poly_order, x_scale=1e-3, y_scale=1e-3, rcond=1e-40)
    logger.info(
        'The time_coa_fit details:\n\t'
        'root mean square residuals = {}\n\t'
        'rank = {}\n\t'
        'singular values = {}'.format(residuals, rank, sing_values))
    return Poly2DType(Coefs=coefs)


def fit_position_xvalidation(
        time_array: numpy.ndarray,
        position_array: numpy.ndarray,
        velocity_array: numpy.ndarray,
        max_degree: int = 5) -> Tuple[
            numpy.ndarray, numpy.ndarray, numpy.ndarray]:
    """
    Empirically fit the polynomials for the X, Y, Z ECF position array, using cross
    validation with the velocity array to determine the best fit degree up to a
    given maximum degree.

    Parameters
    ----------
    time_array : numpy.ndarray
    position_array : numpy.ndarray
    velocity_array : numpy.ndarray
    max_degree : int

    Returns
    -------
    X_poly : numpy.ndarray
    Y_poly : numpy.ndarray
    Z_poly : numpy.ndarray
    """

    if not isinstance(time_array, numpy.ndarray) or \
            not isinstance(position_array, numpy.ndarray) or \
            not isinstance(velocity_array, numpy.ndarray):
        raise TypeError('time_array, position_array, and velocity_array must be numpy.ndarray instances.')

    if time_array.ndim != 1 and time_array.size > 1:
        raise ValueError('time_array must be one-dimensional with at least 2 elements.')

    if position_array.shape != velocity_array.shape:
        raise ValueError('position_array and velocity_array must have the same shape.')

    if position_array.shape[0] != time_array.size:
        raise ValueError('The first dimension of position_array must be the same size as time_array.')

    if position_array.shape[1] != 3:
        raise ValueError('The second dimension of position array must have size 3, '
                         'representing X, Y, Z ECF coordinates.')

    max_degree = int(max_degree)
    if max_degree < 1:
        raise ValueError('max_degree must be at least 1.')
    if max_degree > 10:
        logger.warning(
            'max_degree greater than 10 for polynomial fitting may lead\n\t'
            'to poorly conditioned (i.e. badly behaved) fit.')

    deg = 1
    prev_vel_error = numpy.inf
    P_x, P_y, P_z = None, None, None
    while deg <= min(max_degree, position_array.shape[0]-1):
        # fit position
        P_x = polynomial.polyfit(time_array, position_array[:, 0], deg=deg)
        P_y = polynomial.polyfit(time_array, position_array[:, 1], deg=deg)
        P_z = polynomial.polyfit(time_array, position_array[:, 2], deg=deg)
        # extract estimated velocities
        vel_est = numpy.hstack(
            (polynomial.polyval(time_array, polynomial.polyder(P_x))[:, numpy.newaxis],
             polynomial.polyval(time_array, polynomial.polyder(P_y))[:, numpy.newaxis],
             polynomial.polyval(time_array, polynomial.polyder(P_z))[:, numpy.newaxis]))
        # check our velocity error
        vel_err = vel_est - velocity_array
        cur_vel_error = numpy.sum((vel_err * vel_err))
        deg += 1
        # stop if the error is not smaller than at the previous step
        if cur_vel_error >= prev_vel_error:
            break
    # noinspection PyTypeChecker
    return P_x, P_y, P_z


def sicd_reader_iterator(
        reader,
        partitions=None,
        polarization=None,
        band=None):
    """
    Provides an iterator over a collection of partitions (tuple of tuple of integer
    indices for the reader) for a sicd type reader object.

    Parameters
    ----------
    reader : SICDTypeReader
    partitions : None|tuple
        The partitions collection. If None, then partitioning from
        `reader.get_sicd_partitions()` will be used.
    polarization : None|str
        The polarization string to match.
    band : None|str
        The band to match.

    Returns
    -------
    Iterator[tuple]
        Yields the partition index, the sicd reader index, and the sicd structure.
    """

    def sicd_match():
        match = True
        if band is not None:
            match &= (this_sicd.get_transmit_band_name() == band)
        if polarization is not None:
            match &= (this_sicd.get_processed_polarization() == polarization)
        return match

    if not isinstance(reader, SICDTypeReader):
        raise TypeError('reader must be an instance of SICDTypeReader. Got type {}'.format(type(reader)))
    if reader.reader_type != "SICD":
        raise ValueError('The provided reader must be of SICD type.')

    if partitions is None:
        # noinspection PyUnresolvedReferences
        partitions = reader.get_sicd_partitions()
    # noinspection PyUnresolvedReferences
    the_sicds = reader.get_sicds_as_tuple()
    for this_partition, entry in enumerate(partitions):
        for this_index in entry:
            this_sicd = the_sicds[this_index]
            if sicd_match():
                yield this_partition, this_index, this_sicd


def get_physical_coordinates(
        the_sicd,
        row_value: Union[int, float, numpy.ndarray],
        col_value: Union[int, float, numpy.ndarray]) -> Tuple[Union[float, numpy.ndarray], Union[float, numpy.ndarray]]:
    """
    Transform from image coordinates to physical coordinates, for polynomial evaluation.

    Parameters
    ----------
    the_sicd : sarpy.io.complex.sicd_elements.SICD.SICDType
    row_value : int|float|numpy.ndarray
    col_value : int|float|numpy.ndarray

    Returns
    -------
    row_coords : float|numpy.ndarray
    col_coords : float|numpy.ndarray
    """

    return get_im_physical_coords(row_value, the_sicd.Grid, the_sicd.ImageData, 'row'), \
        get_im_physical_coords(col_value, the_sicd.Grid, the_sicd.ImageData, 'col')


###################
# helper functions

def get_fetch_block_size(
        start_element: int,
        stop_element: int,
        block_size_in_bytes: int,
        bands: int = 1):
    """
    Gets the appropriate block size, given fetch parameters and constraints.

    Note: this is a helper function, and no checking of argument validity will
    be performed.

    Parameters
    ----------
    start_element : int
    stop_element : int
    block_size_in_bytes : int
    bands : int

    Returns
    -------
    int
    """

    if stop_element == start_element:
        return None
    if block_size_in_bytes is None:
        return None

    full_size = float(abs(stop_element - start_element))
    return max(1, int(numpy.ceil(block_size_in_bytes / float(bands*8*full_size))))


def extract_blocks(
        the_range: Tuple[int, int, int],
        index_block_size: Union[None, int, float]) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int]]]:
    """
    Convert the single range definition into a series of range definitions in
    keeping with fetching of the appropriate block sizes.

    Note: this is a helper function, and no checking of argument validity will
    be performed.

    Parameters
    ----------
    the_range : Tuple[int, int, int]
        The input (off processing axis) range.
    index_block_size : None|int|float
        The size of blocks (number of indices).

    Returns
    -------
    range_def: List[Tuple[int, int, int]]
        The sequence of range definitions `(start index, stop index, step)`
        relative to the overall image.
    index_limits: List[Tuple[int, int]]
        The sequence of start/stop indices for positioning of the given range
        relative to the original range.
    """

    entries = numpy.arange(the_range[0], the_range[1], the_range[2], dtype=numpy.int64)
    if index_block_size is None:
        return [the_range, ], [(0, entries.size), ]

    # how many blocks?
    block_count = int(numpy.ceil(entries.size/float(index_block_size)))
    if index_block_size == 1:
        return [the_range, ], [(0, entries.size), ]

    # workspace for what the blocks are
    out1 = []
    out2 = []
    start_ind = 0
    for i in range(block_count):
        end_ind = start_ind+index_block_size
        if end_ind < entries.size:
            block1 = (int(entries[start_ind]), int(entries[end_ind]), the_range[2])
            block2 = (start_ind, end_ind)
        else:
            block1 = (int(entries[start_ind]), the_range[1], the_range[2])
            block2 = (start_ind, entries.size)
        out1.append(block1)
        out2.append(block2)
        start_ind = end_ind
    return out1, out2


def get_data_mean_magnitude(bounds, reader, index, block_size_in_bytes):
    """
    Gets the mean magnitude in the region defined by bounds.

    Note: this is a helper function, and no checking of argument validity will
    be performed.

    Parameters
    ----------
    bounds : numpy.ndarray|tuple|list
        Of the form `(row_start, row_end, col_start, col_end)`.
    reader : SICDTypeReader
        The data reader.
    index : int
        The reader index to use.
    block_size_in_bytes : int|float
        The block size in bytes.

    Returns
    -------
    float
    """

    # Extract the mean of the data magnitude - for global remap usage
    logger.info(
        'Calculating mean over the block ({}:{}, {}:{}), this may be time consuming'.format(*bounds))
    mean_block_size = get_fetch_block_size(bounds[0], bounds[1], block_size_in_bytes)
    mean_column_blocks, _ = extract_blocks((bounds[2], bounds[3], 1), mean_block_size)
    mean_total = 0.0
    mean_count = 0
    # noinspection PyTypeChecker
    for this_column_range in mean_column_blocks:
        data = numpy.abs(reader[
                         bounds[0]:bounds[1],
                         this_column_range[0]:this_column_range[1],
                         index])
        mask = (data > 0) & numpy.isfinite(data)
        mean_total += numpy.sum(data[mask])
        mean_count += numpy.sum(mask)
    return float(mean_total / mean_count)


def stats_calculation(data, percentile=None):
    """
    Calculate the statistics for input into the nrl remap.

    Parameters
    ----------
    data : numpy.ndarray
        The amplitude array, assumed real valued an finite.
    percentile : None|float|int
        Which percentile to calculate

    Returns
    -------
    tuple
        Of the form `(`minimum, maximum)` or `(minimum, maximum, `percentile` percentile)
    """

    if percentile is None:
        return numpy.min(data), numpy.max(data)
    else:
        return numpy.min(data), numpy.max(data), scoreatpercentile(data, percentile)


def get_data_extrema(bounds, reader, index, block_size_in_bytes, percentile=None):
    """
    Gets the minimum and maximum magnitude in the region defined by bounds,
    optionally, get as **estimate** for the `percentage` percentile point.

    Note: this is a helper function, and no checking of argument validity will
    be performed.

    Parameters
    ----------
    bounds : numpy.ndarray|tuple|list
        Of the form `(row_start, row_end, col_start, col_end)`.
    reader : SICDTypeReader
        The data reader.
    index : int
        The reader index to use.
    block_size_in_bytes : int|float
        The block size in bytes.
    percentile : None|int|float
        The optional percentile to estimate.

    Returns
    -------
    Tuple
        The minimum (finite) observed, maximum (finite) observed, and optionally,
        the desired (of finite) percentile.

        The minimum, maximum, and percentile values will be `None` exactly when the reader
        contains no finite data.
    """

    min_value = None
    max_value = None
    percent = None

    # Extract the mean of the data magnitude - for global remap usage
    logger.info(
        'Calculating extrema over the block ({}:{}, {}:{}), this may be time consuming'.format(*bounds))
    mean_block_size = get_fetch_block_size(bounds[0], bounds[1], block_size_in_bytes)
    mean_column_blocks, _ = extract_blocks((bounds[2], bounds[3], 1), mean_block_size)
    # noinspection PyTypeChecker
    for this_column_range in mean_column_blocks:
        data = numpy.abs(reader[
                         bounds[0]:bounds[1],
                         this_column_range[0]:this_column_range[1],
                         index])
        mask = numpy.isfinite(data)
        if numpy.any(mask):
            temp_values = stats_calculation(data[mask], percentile=percentile)

            min_value = temp_values[0] if min_value is None else min(min_value, temp_values[0])
            max_value = temp_values[1] if max_value is None else max(max_value, temp_values[1])
            percent = temp_values[2] if percentile is not None else None

    if percentile is None:
        return min_value, max_value
    else:
        return min_value, max_value, percent



"""
Basic building blocks for CPHD standard - mostly overlap with SICD elements
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Union, List, Dict

import numpy

from sarpy.io.xml.base import Serializable, Arrayable, SerializableArray, ParametersCollection
from sarpy.io.xml.descriptors import SerializableDescriptor, SerializableArrayDescriptor, \
    IntegerDescriptor, FloatDescriptor, ParametersDescriptor

from .base import DEFAULT_STRICT


_len2_array_text = 'Expected array to be of length 2,\n\t' \
                   'and received `{}`'
_array_type_text = 'Expected array to be numpy.ndarray, list, or tuple,\n\tgot `{}`'

###################
# module variables
POLARIZATION_TYPE = ('H', 'V', 'X', 'Y', 'S', 'E', 'RHC', 'LHC', 'UNSPECIFIED')


class LSType(Serializable, Arrayable):
    """
    Represents line and sample.
    """

    _fields = ('Line', 'Sample')
    _required = _fields
    _numeric_format = {'Line': '0.17G', 'Sample': '0.17G'}
    # Descriptor
    Line = FloatDescriptor(
        'Line', _required, strict=DEFAULT_STRICT,
        docstring='The Line.')  # type: float
    Sample = FloatDescriptor(
        'Sample', _required, strict=DEFAULT_STRICT,
        docstring='The Sample.')  # type: float

    def __init__(self, Line=None, Sample=None, **kwargs):
        """

        Parameters
        ----------
        Line : float
        Sample : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Line = Line
        self.Sample = Sample
        super(LSType, self).__init__(**kwargs)

    def get_array(self, dtype=numpy.float64):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [Line, Sample]
        """

        return numpy.array([self.Line, self.Sample], dtype=dtype)

    @classmethod
    def from_array(cls, array):
        """
        Construct from an iterable.

        Parameters
        ----------
        array : numpy.ndarray|list|tuple

        Returns
        -------
        LSType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) != 2:
                raise ValueError(_len2_array_text.format(array))
            return cls(Line=array[0], Sample=array[1])
        raise ValueError(_array_type_text.format(type(array)))


class LSVertexType(LSType):
    """
    An array element of LSType.
    """

    _fields = ('Line', 'Sample', 'index')
    _required = _fields
    _set_as_attribute = ('index', )
    # descriptors
    index = IntegerDescriptor(
        'index', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='The array index.')  # type: int

    def __init__(self, Line=None, Sample=None, index=None, **kwargs):
        """

        Parameters
        ----------
        Line : float
        Sample : float
        index : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.index = index
        super(LSVertexType, self).__init__(Line=Line, Sample=Sample, **kwargs)

    @classmethod
    def from_array(cls, array, index=1):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [Line, Sample]
        index : int
            array index
        Returns
        -------
        XYVertexType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) != 2:
                raise ValueError(_len2_array_text.format(array))
            return cls(Line=array[0], Sample=array[1], index=index)
        raise ValueError(_array_type_text.format(type(array)))


class XYType(Serializable, Arrayable):
    """
    A point in two-dimensional spatial coordinates.
    """

    _fields = ('X', 'Y')
    _required = _fields
    _numeric_format = {'X': '0.17G', 'Y': '0.17G'}
    # descriptors
    X = FloatDescriptor(
        'X', _required, strict=True,
        docstring='The X attribute. Assumed to ECF or other, similar '
                  'coordinates.')  # type: float
    Y = FloatDescriptor(
        'Y', _required, strict=True,
        docstring='The Y attribute. Assumed to ECF or other, similar '
                  'coordinates.')  # type: float

    def __init__(self, X=None, Y=None, **kwargs):
        """
        Parameters
        ----------
        X : float
        Y : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.X, self.Y = X, Y
        super(XYType, self).__init__(**kwargs)

    @classmethod
    def from_array(cls, array):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [X, Y]

        Returns
        -------
        XYType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) != 2:
                raise ValueError(_len2_array_text.format(array))
            return cls(X=array[0], Y=array[1])
        raise ValueError(_array_type_text.format(type(array)))

    def get_array(self, dtype=numpy.float64):
        """
        Gets an array representation of the class instance.

        Parameters
        ----------
        dtype : str|numpy.dtype|numpy.number
            numpy data type of the return

        Returns
        -------
        numpy.ndarray
            array of the form [X,Y]
        """

        return numpy.array([self.X, self.Y], dtype=dtype)


class XYVertexType(XYType):
    """
    An array element of XYType.
    """

    _fields = ('X', 'Y', 'index')
    _required = _fields
    _set_as_attribute = ('index', )
    # descriptors
    index = IntegerDescriptor(
        'index', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='The array index.')  # type: int

    def __init__(self, X=None, Y=None, index=None, **kwargs):
        """

        Parameters
        ----------
        X : float
        Y : float
        index : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.index = index
        super(XYVertexType, self).__init__(X=X, Y=Y, **kwargs)

    @classmethod
    def from_array(cls, array, index=1):
        """
        Create from an array type entry.

        Parameters
        ----------
        array: numpy.ndarray|list|tuple
            assumed [X, Y]
        index : int
            array index
        Returns
        -------
        XYVertexType
        """

        if array is None:
            return None
        if isinstance(array, (numpy.ndarray, list, tuple)):
            if len(array) != 2:
                raise ValueError(_len2_array_text.format(array))
            return cls(X=array[0], Y=array[1], index=index)
        raise ValueError(_array_type_text.format(type(array)))


class AreaType(Serializable):
    """
    An area object.
    """

    _fields = ('X1Y1', 'X2Y2', 'Polygon')
    _required = ('X1Y1', 'X2Y2')
    _collections_tags = {'Polygon': {'array': True, 'child_tag': 'Vertex'}}
    # descriptors
    X1Y1 = SerializableDescriptor(
        'X1Y1', XYType, _required, strict=DEFAULT_STRICT,
        docstring='*"Minimum"* corner of the rectangle in Image '
                  'coordinates.')  # type: XYType
    X2Y2 = SerializableDescriptor(
        'X2Y2', XYType, _required, strict=DEFAULT_STRICT,
        docstring='*"Maximum"* corner of the rectangle in Image '
                  'coordinates.')  # type: XYType
    Polygon = SerializableArrayDescriptor(
        'Polygon', XYVertexType, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=3,
        docstring='Polygon further reducing the bounding box, in Image '
                  'coordinates.')  # type: Union[SerializableArray, List[XYVertexType]]

    def __init__(self, X1Y1=None, X2Y2=None, Polygon=None, **kwargs):
        """

        Parameters
        ----------
        X1Y1 : XYType|numpy.ndarray|list|tuple
        X2Y2 : XYType|numpy.ndarray|list|tuple
        Polygon : SerializableArray|List[XYVertexType]|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.X1Y1 = X1Y1
        self.X2Y2 = X2Y2
        self.Polygon = Polygon
        super(AreaType, self).__init__(**kwargs)


class AddedParametersType(Serializable):
    _collections_tags = {
        'Parameters': {'array': False, 'child_tag': 'Parameter'},
    }
    _fields = ('Parameters',)
    _required = _fields

    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Free form parameters object collection.')  # type: ParametersCollection

    def __init__(
            self,
            Parameters: Union[ParametersCollection, Dict] = None,
            **kwargs):
        """

        Parameters
        ----------
        Parameters : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Parameters = Parameters
        super().__init__(**kwargs)



"""
Functionality for reading Capella SAR data into a SICD model.

**This functionality is really onl partially complete**
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Wade Schwartzkopf")


import logging
import json
from typing import Dict, Any, Tuple, Union, Optional
from collections import OrderedDict

from scipy.constants import speed_of_light
import numpy
from numpy.polynomial import polynomial

from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.tiff import TiffDetails, NativeTiffDataSegment
from sarpy.io.general.utils import parse_timestring, get_seconds, is_file_like
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.utils import fit_position_xvalidation
from sarpy.io.complex.sicd_elements.blocks import XYZPolyType, Poly2DType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType, RadarModeType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType
from sarpy.io.complex.sicd_elements.GeoData import GeoDataType, SCPType
from sarpy.io.complex.sicd_elements.Position import PositionType
from sarpy.io.complex.sicd_elements.Grid import GridType, DirParamType, WgtTypeType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType, \
    WaveformParametersType, ChanParametersType
from sarpy.io.complex.sicd_elements.Timeline import TimelineType, IPPSetType
from sarpy.io.complex.sicd_elements.ImageFormation import ImageFormationType, \
    RcvChanProcType, ProcessingType
from sarpy.io.complex.sicd_elements.RMA import RMAType, INCAType
from sarpy.io.complex.sicd_elements.Radiometric import RadiometricType, NoiseLevelType_

logger = logging.getLogger(__name__)


#########
# helper functions

def avci_nacaroglu_window(M, alpha=1.25):
    """
    Avci-Nacaroglu Exponential window. See Doerry '17 paper window 4.40 p 154
    Parameters
    ----------
    M : int
    alpha : float
    """

    M2 = 0.5*M
    t = (numpy.arange(M) - M2)/M
    return numpy.exp(numpy.pi*alpha*(numpy.sqrt(1 - (2*t)**2) - 1))


###########
# parser and interpreter for tiff attributes

class CapellaDetails(object):
    """
    Parses and converts the Cosmo Skymed metadata
    """

    __slots__ = ('_tiff_details', '_img_desc_tags')

    def __init__(self, file_name: str):
        """

        Parameters
        ----------
        file_name : str
        """

        # verify that the file is a tiff file
        self._tiff_details = TiffDetails(file_name)
        # verify that ImageDescription tiff tag exists
        if 'ImageDescription' not in self._tiff_details.tags:
            raise SarpyIOError('No "ImageDescription" tag in the tiff.')

        img_format = self._tiff_details.tags['ImageDescription']
        # verify that ImageDescription has a reasonable format
        try:
            self._img_desc_tags = json.loads(img_format)  # type: Dict[str, Any]
        except Exception as e:
            msg = 'Failed deserializing the ImageDescription tag as json with error {}'.format(e)
            logger.info(msg)
            raise SarpyIOError(msg)

        # verify the file is not compressed
        self._tiff_details.check_compression()
        # verify the file is not tiled
        self._tiff_details.check_tiled()

    @property
    def file_name(self) -> str:
        """
        str: the file name
        """

        return self._tiff_details.file_name

    @property
    def tiff_details(self) -> TiffDetails:
        """
        TiffDetails: The tiff details object.
        """

        return self._tiff_details

    def get_symmetry(self) -> (Union[None, Tuple[int, ...]], Tuple[int, ...]):
        """
        Gets the symmetry operations definition.

        Returns
        -------
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : Tuple[int, ...]
        """

        pointing = self._img_desc_tags['collect']['radar']['pointing'].lower()
        if pointing == 'left':
            return (0, ), (1, 0, 2)
        elif pointing == 'right':
            return None, (1, 0, 2)
        else:
            raise ValueError('Got unhandled pointing value {}'.format(pointing))

    def get_sicd(self) -> SICDType:
        """
        Get the SICD metadata for the image.

        Returns
        -------
        SICDType
        """

        def convert_string_dict(dict_in: dict) -> dict:
            dict_out = OrderedDict()
            for key, val in dict_in.items():
                if isinstance(val, str):
                    dict_out[key] = val
                elif isinstance(val, int):
                    dict_out[key] = str(val)
                elif isinstance(val, float):
                    dict_out[key] = '{0:0.17G}'.format(val)
                else:
                    raise TypeError('Got unhandled type {}'.format(type(val)))
            return dict_out

        def extract_state_vector() -> (numpy.ndarray, numpy.ndarray, numpy.ndarray):
            vecs = collect['state']['state_vectors']
            times = numpy.zeros((len(vecs), ), dtype=numpy.float64)
            positions = numpy.zeros((len(vecs), 3), dtype=numpy.float64)
            velocities = numpy.zeros((len(vecs), 3), dtype=numpy.float64)
            for i, entry in enumerate(vecs):
                times[i] = get_seconds(parse_timestring(entry['time'], precision='ns'), start_time, precision='ns')
                positions[i, :] = entry['position']
                velocities[i, :] = entry['velocity']
            return times, positions, velocities

        def get_radar_parameter(name) -> Any:
            if name in radar:
                return radar[name]
            if len(radar_time_varying) > 0:
                element = radar_time_varying[0]
                if name in element:
                    return element[name]
            raise ValueError('Unable to determine radar parameter `{}`'.format(name))

        def get_collection_info() -> CollectionInfoType:
            coll_name = collect['platform']
            mode = collect['mode'].strip().lower()
            if mode == 'stripmap':
                radar_mode = RadarModeType(ModeType='STRIPMAP', ModeID=mode)
            elif mode == 'spotlight':
                radar_mode = RadarModeType(ModeType='SPOTLIGHT', ModeID=mode)
            elif mode == 'sliding_spotlight':
                radar_mode = RadarModeType(ModeType='DYNAMIC STRIPMAP', ModeID=mode)
            else:
                raise ValueError('Got unhandled radar mode {}'.format(mode))

            return CollectionInfoType(
                CollectorName=coll_name,
                CoreName=collect['collect_id'],
                RadarMode=radar_mode,
                Classification='UNCLASSIFIED',
                CollectType='MONOSTATIC')

        def get_image_creation() -> ImageCreationType:
            from sarpy.__about__ import __version__
            return ImageCreationType(
                Application=self._tiff_details.tags['Software'],
                DateTime=parse_timestring(self._img_desc_tags['processing_time'], precision='us'),
                Profile='sarpy {}'.format(__version__),
                Site='Unknown')

        def get_image_data() -> ImageDataType:
            rows = int(img['columns'])  # capella uses flipped row/column definition?
            cols = int(img['rows'])
            if img['data_type'] == 'CInt16':
                pixel_type = 'RE16I_IM16I'
            else:
                raise ValueError('Got unhandled data_type {}'.format(img['data_type']))

            scp_pixel = (int(0.5 * rows), int(0.5 * cols))
            if radar['pointing'] == 'left':
                scp_pixel = (rows - scp_pixel[0] - 1, cols - scp_pixel[1] - 1)

            return ImageDataType(
                NumRows=rows,
                NumCols=cols,
                FirstRow=0,
                FirstCol=0,
                PixelType=pixel_type,
                FullImage=(rows, cols),
                SCPPixel=scp_pixel)

        def get_geo_data() -> GeoDataType:
            return GeoDataType(SCP=SCPType(ECF=img['center_pixel']['target_position']))

        def get_position() -> PositionType:
            px, py, pz = fit_position_xvalidation(state_time, state_position, state_velocity, max_degree=8)
            return PositionType(ARPPoly=XYZPolyType(X=px, Y=py, Z=pz))

        def get_grid() -> GridType:

            def get_weight(window_dict: dict) -> (WgtTypeType, Union[None, numpy.ndarray]):
                window_name = window_dict['name']
                if window_name.lower() == 'rectangular':
                    return WgtTypeType(WindowName='UNIFORM'), None
                elif window_name.lower() == 'avci-nacaroglu':
                    return WgtTypeType(
                        WindowName=window_name.upper(),
                        Parameters=convert_string_dict(window_dict['parameters'])), \
                           avci_nacaroglu_window(64, alpha=window_dict['parameters']['alpha'])
                else:
                    return WgtTypeType(
                        WindowName=window_name,
                        Parameters=convert_string_dict(window_dict['parameters'])), None

            image_plane = 'SLANT'
            grid_type = 'RGZERO'

            coa_time = parse_timestring(img['center_pixel']['center_time'], precision='ns')
            row_bw = img.get('processed_range_bandwidth', bw)
            row_imp_rsp_bw = 2*row_bw/speed_of_light
            row_wgt, row_wgt_funct = get_weight(img['range_window'])
            row = DirParamType(
                SS=img['image_geometry']['delta_range_sample'],
                Sgn=-1,
                ImpRespBW=row_imp_rsp_bw,
                ImpRespWid=img['range_resolution'],
                KCtr=2*fc/speed_of_light,
                DeltaK1=-0.5*row_imp_rsp_bw,
                DeltaK2=0.5*row_imp_rsp_bw,
                DeltaKCOAPoly=[[0.0, ], ],
                WgtFunct=row_wgt_funct,
                WgtType=row_wgt)

            # get timecoa value
            timecoa_value = get_seconds(coa_time, start_time)
            # find an approximation for zero doppler spacing - necessarily rough for backprojected images
            col_ss = img['pixel_spacing_row']
            dop_bw = img['processed_azimuth_bandwidth']

            col_wgt, col_wgt_funct = get_weight(img['azimuth_window'])
            col = DirParamType(
                SS=col_ss,
                Sgn=-1,
                ImpRespWid=img['azimuth_resolution'],
                ImpRespBW=dop_bw*abs(ss_zd_s)/col_ss,
                KCtr=0,
                WgtFunct=col_wgt_funct,
                WgtType=col_wgt)

            # TODO:
            #   column deltakcoa poly - it's in there at ["image"]["frequency_doppler_centroid_polynomial"]

            return GridType(
                ImagePlane=image_plane,
                Type=grid_type,
                TimeCOAPoly=[[timecoa_value, ], ],
                Row=row,
                Col=col)

        def get_radar_collection():
            # type: () -> RadarCollectionType

            freq_min = fc - 0.5*bw
            return RadarCollectionType(
                TxPolarization=radar['transmit_polarization'],
                TxFrequency=(freq_min, freq_min + bw),
                Waveform=[WaveformParametersType(
                    TxRFBandwidth=bw,
                    TxPulseLength=get_radar_parameter('pulse_duration'),
                    RcvDemodType='CHIRP',
                    ADCSampleRate=radar['sampling_frequency'],
                    TxFreqStart=freq_min)],
                RcvChannels=[ChanParametersType(
                    TxRcvPolarization='{}:{}'.format(radar['transmit_polarization'],
                                                     radar['receive_polarization']))])

        def get_timeline() -> TimelineType:
            prf = radar['prf'][0]['prf']
            return TimelineType(
                CollectStart=start_time,
                CollectDuration=duration,
                IPP=[
                    IPPSetType(
                        TStart=0,
                        TEnd=duration,
                        IPPStart=0,
                        IPPEnd=round(duration*prf) - 1,
                        IPPPoly=(0, prf)), ])

        def get_image_formation() -> ImageFormationType:
            algo = img['algorithm'].upper()
            processings = None
            if algo == 'BACKPROJECTION':
                processings = [ProcessingType(Type='Backprojected to DEM', Applied=True), ]
            else:
                logger.warning(
                    'Got unexpected algorithm, the results for the '
                    'sicd structure might be unexpected')

            if algo not in ('PFA', 'RMA', 'RGAZCOMP'):
                logger.warning(
                    'Image formation algorithm {} not one of the recognized SICD options, '
                    'being set to "OTHER".'.format(algo))
                algo = 'OTHER'

            return ImageFormationType(
                RcvChanProc=RcvChanProcType(NumChanProc=1, PRFScaleFactor=1),
                ImageFormAlgo=algo,
                TStartProc=0,
                TEndProc=duration,
                TxRcvPolarizationProc='{}:{}'.format(radar['transmit_polarization'], radar['receive_polarization']),
                TxFrequencyProc=(
                    radar_collection.TxFrequency.Min,
                    radar_collection.TxFrequency.Max),
                STBeamComp='NO',
                ImageBeamComp='NO',
                AzAutofocus='NO',
                RgAutofocus='NO',
                Processings=processings)

        def get_rma() -> RMAType:
            img_geometry = img['image_geometry']
            near_range = img_geometry['range_to_first_sample']
            center_time = parse_timestring(img['center_pixel']['center_time'], precision='us')
            first_time = parse_timestring(img_geometry['first_line_time'], precision='us')
            zd_time_scp = get_seconds(center_time, first_time, 'us')
            r_ca_scp = near_range + image_data.SCPPixel.Row*grid.Row.SS
            time_ca_poly = numpy.array([zd_time_scp, -look*ss_zd_s/grid.Col.SS], dtype='float64')

            timecoa_value = get_seconds(center_time, start_time)
            arp_velocity = position.ARPPoly.derivative_eval(timecoa_value, der_order=1)
            vm_ca = numpy.linalg.norm(arp_velocity)
            inca = INCAType(
                R_CA_SCP=r_ca_scp,
                FreqZero=fc,
                TimeCAPoly=time_ca_poly,
                DRateSFPoly=[[1/(vm_ca*ss_zd_s/grid.Col.SS)], ]
            )

            return RMAType(
                RMAlgoType='RG_DOP',
                INCA=inca)

        def get_radiometric() -> Union[None, RadiometricType]:
            if img['radiometry'].lower() != 'beta_nought':
                logger.warning(
                    'Got unrecognized Capella radiometry {},\n\t'
                    'skipping the radiometric metadata'.format(img['radiometry']))
                return None

            return RadiometricType(BetaZeroSFPoly=[[img['scale_factor']**2, ], ])

        def add_noise() -> None:
            if sicd.Radiometric is None:
                return

            nesz_raw = numpy.array(img['nesz_polynomial']['coefficients'], dtype='float64')
            test_value = polynomial.polyval(rma.INCA.R_CA_SCP, nesz_raw)
            if abs(test_value - img['nesz_peak']) > 100:
                # this polynomial reversed in early versions, so reverse if evaluated results are nonsense
                nesz_raw = nesz_raw[::-1]
            nesz_poly_raw = Poly2DType(Coefs=numpy.reshape(nesz_raw, (-1, 1)))
            noise_coeffs = nesz_poly_raw.shift(-rma.INCA.R_CA_SCP, 1, 0, 1, return_poly=False)
            # this is in nesz units, so shift to absolute units
            noise_coeffs[0] -= 10*numpy.log10(sicd.Radiometric.SigmaZeroSFPoly[0, 0])
            sicd.Radiometric.NoiseLevel = NoiseLevelType_(NoiseLevelType='ABSOLUTE', NoisePoly=noise_coeffs)

        # extract general use information
        collect = self._img_desc_tags['collect']
        img = collect['image']
        radar = collect['radar']
        radar_time_varying = radar.get('time_varying_parameters', [])

        start_time = parse_timestring(collect['start_timestamp'], precision='ns')
        end_time = parse_timestring(collect['stop_timestamp'], precision='ns')
        duration = get_seconds(end_time, start_time, precision='ns')
        state_time, state_position, state_velocity = extract_state_vector()
        bw = get_radar_parameter('pulse_bandwidth')
        fc = get_radar_parameter('center_frequency')
        ss_zd_s = img['image_geometry']['delta_line_time']
        look = -1 if radar['pointing'] == 'right' else 1

        # define the sicd elements
        collection_info = get_collection_info()
        image_creation = get_image_creation()
        image_data = get_image_data()
        geo_data = get_geo_data()
        position = get_position()
        grid = get_grid()
        radar_collection = get_radar_collection()
        timeline = get_timeline()
        image_formation = get_image_formation()
        rma = get_rma()
        radiometric = get_radiometric()

        sicd = SICDType(
            CollectionInfo=collection_info,
            ImageCreation=image_creation,
            ImageData=image_data,
            GeoData=geo_data,
            Position=position,
            Grid=grid,
            RadarCollection=radar_collection,
            Timeline=timeline,
            ImageFormation=image_formation,
            RMA=rma,
            Radiometric=radiometric)
        sicd.derive()

        add_noise()
        sicd.populate_rniirs(override=False)
        return sicd


class CapellaReader(SICDTypeReader):
    """
    The Capella SLC reader implementation. **This is only partially complete.**

    **Changed in version 1.3.0** for reading changes.
    """

    __slots__ = ('_capella_details', )

    def __init__(self, capella_details):
        """

        Parameters
        ----------
        capella_details : str|CapellaDetails
        """

        if isinstance(capella_details, str):
            capella_details = CapellaDetails(capella_details)

        if not isinstance(capella_details, CapellaDetails):
            raise TypeError('The input argument for capella_details must be a '
                            'filename or CapellaDetails object')
        self._capella_details = capella_details
        sicd = self.capella_details.get_sicd()
        reverse_axes, transpose_axes = self.capella_details.get_symmetry()
        data_segment = NativeTiffDataSegment(
            self.capella_details.tiff_details, reverse_axes=reverse_axes, transpose_axes=transpose_axes)

        SICDTypeReader.__init__(self, data_segment, sicd, close_segments=True)
        self._check_sizes()

    @property
    def capella_details(self) -> CapellaDetails:
        """
        CapellaDetails: The capella details object.
        """

        return self._capella_details

    @property
    def file_name(self) -> str:
        return self.capella_details.file_name


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name: str) -> Optional[CapellaReader]:
    """
    Tests whether a given file_name corresponds to a Capella SAR file.
    Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    CapellaReader|None
        `CapellaReader` instance if Capella file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    try:
        capella_details = CapellaDetails(file_name)
        logger.info('File {} is determined to be a Capella file.'.format(file_name))
        return CapellaReader(capella_details)
    except SarpyIOError:
        return None




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PT(TREElement):
    def __init__(self, value):
        super(PT, self).__init__()
        self.add_field('PID', 's', 10, value)
        self.add_field('LON', 's', 15, value)
        self.add_field('LAT', 's', 15, value)
        self.add_field('ZVL', 's', 15, value)
        self.add_field('DIX', 's', 11, value)
        self.add_field('DIY', 's', 11, value)


class REGPTBType(TREElement):
    def __init__(self, value):
        super(REGPTBType, self).__init__()
        self.add_field('NUMPTS', 'd', 4, value)
        self.add_loop('PTs', self.NUMPTS, PT, value)


class REGPTB(TREExtension):
    _tag_value = 'REGPTB'
    _data_type = REGPTBType
